@@ -77,20 +77,20 @@ static void serialize(MVMThreadContext *tc, MVMSTable *st, void *data, MVMSerial
     MVM_serialization_write_ref(tc, writer, body->fates);
 
     /* Write number of states. */
-    MVM_serialization_write_varint(tc, writer, body->num_states);
+    MVM_serialization_write_int(tc, writer, body->num_states);
 
     /* Write state edge list counts. */
     for (i = 0; i < body->num_states; i++)
-        MVM_serialization_write_varint(tc, writer, body->num_state_edges[i]);
+        MVM_serialization_write_int(tc, writer, body->num_state_edges[i]);
 
     /* Write state graph. */
     for (i = 0; i < body->num_states; i++) {
         for (j = 0; j < body->num_state_edges[i]; j++) {
-            MVM_serialization_write_varint(tc, writer, body->states[i][j].act);
-            MVM_serialization_write_varint(tc, writer, body->states[i][j].to);
+            MVM_serialization_write_int(tc, writer, body->states[i][j].act);
+            MVM_serialization_write_int(tc, writer, body->states[i][j].to);
             switch (body->states[i][j].act & 0xff) {
                 case MVM_NFA_EDGE_FATE:
-                    MVM_serialization_write_varint(tc, writer, body->states[i][j].arg.i);
+                    MVM_serialization_write_int(tc, writer, body->states[i][j].arg.i);
                     break;
                 case MVM_NFA_EDGE_CODEPOINT:
                 case MVM_NFA_EDGE_CODEPOINT_LL:
@@ -100,23 +100,23 @@ static void serialize(MVMThreadContext *tc, MVMSTable *st, void *data, MVMSerial
                     MVMGrapheme32 g = body->states[i][j].arg.g;
                     if (g >= 0) {
                         /* Non-synthetic. */
-                        MVM_serialization_write_varint(tc, writer, g);
+                        MVM_serialization_write_int(tc, writer, g);
                     }
                     else {
                         /* Synthetic. Write the number of codepoints negated,
                          * and then each of the codepoints. */
                         MVMNFGSynthetic *si = MVM_nfg_get_synthetic_info(tc, g);
                         MVMint32 k;
-                        MVM_serialization_write_varint(tc, writer, -(si->num_combs + 1));
-                        MVM_serialization_write_varint(tc, writer, si->base);
+                        MVM_serialization_write_int(tc, writer, -(si->num_combs + 1));
+                        MVM_serialization_write_int(tc, writer, si->base);
                         for (k = 0; k < si->num_combs; k++)
-                            MVM_serialization_write_varint(tc, writer, si->combs[k]);
+                            MVM_serialization_write_int(tc, writer, si->combs[k]);
                     }
                     break;
                 }
                 case MVM_NFA_EDGE_CHARCLASS:
                 case MVM_NFA_EDGE_CHARCLASS_NEG:
-                    MVM_serialization_write_varint(tc, writer, body->states[i][j].arg.i);
+                    MVM_serialization_write_int(tc, writer, body->states[i][j].arg.i);
                     break;
                 case MVM_NFA_EDGE_CHARLIST:
                 case MVM_NFA_EDGE_CHARLIST_NEG:
@@ -131,8 +131,8 @@ static void serialize(MVMThreadContext *tc, MVMSTable *st, void *data, MVMSerial
                 case MVM_NFA_EDGE_CHARRANGE_NEG:
                 case MVM_NFA_EDGE_CHARRANGE_M:
                 case MVM_NFA_EDGE_CHARRANGE_M_NEG: {
-                    MVM_serialization_write_varint(tc, writer, body->states[i][j].arg.uclc.lc);
-                    MVM_serialization_write_varint(tc, writer, body->states[i][j].arg.uclc.uc);
+                    MVM_serialization_write_int(tc, writer, body->states[i][j].arg.uclc.lc);
+                    MVM_serialization_write_int(tc, writer, body->states[i][j].arg.uclc.uc);
                     break;
                 }
             }
@@ -149,13 +149,13 @@ static void deserialize(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, vo
     body->fates = MVM_serialization_read_ref(tc, reader);
 
     /* Read number of states. */
-    body->num_states = MVM_serialization_read_varint(tc, reader);
+    body->num_states = MVM_serialization_read_int(tc, reader);
 
     if (body->num_states > 0) {
         /* Read state edge list counts. */
         body->num_state_edges = MVM_malloc(body->num_states * sizeof(MVMint64));
         for (i = 0; i < body->num_states; i++)
-            body->num_state_edges[i] = MVM_serialization_read_varint(tc, reader);
+            body->num_state_edges[i] = MVM_serialization_read_int(tc, reader);
 
         /* Read state graph. */
         body->states = MVM_malloc(body->num_states * sizeof(MVMNFAStateInfo *));
@@ -164,18 +164,18 @@ static void deserialize(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, vo
             if (edges > 0)
                 body->states[i] = MVM_malloc(edges * sizeof(MVMNFAStateInfo));
             for (j = 0; j < edges; j++) {
-                body->states[i][j].act = MVM_serialization_read_varint(tc, reader);
-                body->states[i][j].to = MVM_serialization_read_varint(tc, reader);
+                body->states[i][j].act = MVM_serialization_read_int(tc, reader);
+                body->states[i][j].to = MVM_serialization_read_int(tc, reader);
                 switch (body->states[i][j].act & 0xff) {
                     case MVM_NFA_EDGE_FATE:
-                        body->states[i][j].arg.i = MVM_serialization_read_varint(tc, reader);
+                        body->states[i][j].arg.i = MVM_serialization_read_int(tc, reader);
                         break;
                     case MVM_NFA_EDGE_CODEPOINT:
                     case MVM_NFA_EDGE_CODEPOINT_LL:
                     case MVM_NFA_EDGE_CODEPOINT_NEG:
                     case MVM_NFA_EDGE_CODEPOINT_M:
                     case MVM_NFA_EDGE_CODEPOINT_M_NEG: {
-                        MVMint64 cp_or_synth_count = MVM_serialization_read_varint(tc, reader);
+                        MVMint64 cp_or_synth_count = MVM_serialization_read_int(tc, reader);
                         if (cp_or_synth_count >= 0) {
                             body->states[i][j].arg.g = (MVMGrapheme32)cp_or_synth_count;
                         }
@@ -184,7 +184,7 @@ static void deserialize(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, vo
                             MVMCodepoint *codes = MVM_malloc(num_codes * sizeof(MVMCodepoint));
                             MVMint32 k;
                             for (k = 0; k < num_codes; k++)
-                                codes[k] = (MVMCodepoint)MVM_serialization_read_varint(tc, reader);
+                                codes[k] = (MVMCodepoint)MVM_serialization_read_int(tc, reader);
                             body->states[i][j].arg.g = MVM_nfg_codes_to_grapheme(tc, codes, num_codes);
                             MVM_free(codes);
                         }
@@ -192,7 +192,7 @@ static void deserialize(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, vo
                     }
                     case MVM_NFA_EDGE_CHARCLASS:
                     case MVM_NFA_EDGE_CHARCLASS_NEG:
-                        body->states[i][j].arg.i = MVM_serialization_read_varint(tc, reader);
+                        body->states[i][j].arg.i = MVM_serialization_read_int(tc, reader);
                         break;
                     case MVM_NFA_EDGE_CHARLIST:
                     case MVM_NFA_EDGE_CHARLIST_NEG:
@@ -207,8 +207,8 @@ static void deserialize(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, vo
                     case MVM_NFA_EDGE_CHARRANGE_NEG:
                     case MVM_NFA_EDGE_CHARRANGE_M:
                     case MVM_NFA_EDGE_CHARRANGE_M_NEG: {
-                        body->states[i][j].arg.uclc.lc = MVM_serialization_read_varint(tc, reader);
-                        body->states[i][j].arg.uclc.uc = MVM_serialization_read_varint(tc, reader);
+                        body->states[i][j].arg.uclc.lc = MVM_serialization_read_int(tc, reader);
+                        body->states[i][j].arg.uclc.uc = MVM_serialization_read_int(tc, reader);
                         break;
                     }
                 }
@@ -272,7 +272,6 @@ static const MVMREPROps this_repr = {
     NULL, /* spesh */
     "NFA", /* name */
     MVM_REPR_ID_NFA,
-    0, /* refs_frames */
     unmanaged_size,
     NULL, /* describe_refs */
 };
@@ -493,7 +492,6 @@ static MVMint64 * nqp_nfa_run(MVMThreadContext *tc, MVMNFABody *nfa, MVMString *
                         MVMint64 arg = edge_info[i].arg.i;
                         MVMint64 j;
                         MVMint64 found_fate = 0;
-                        arg &= 0xffffff;   /* XXX can go away after rebootstrap */
                         if (nfadeb)
                             fprintf(stderr, "fate(%016llx) ", (long long unsigned int)arg);
                         for (j = 0; j < total_fates; j++) {
@@ -662,7 +660,7 @@ static MVMint64 * nqp_nfa_run(MVMThreadContext *tc, MVMNFABody *nfa, MVMString *
                             if (((act == MVM_NFA_EDGE_CODEPOINT_M)     && (ga == gb))
                              || ((act == MVM_NFA_EDGE_CODEPOINT_M_NEG) && (ga != gb)))
                                 nextst[numnext++] = to;
-
+                            MVM_unicode_normalizer_cleanup(tc, &norm);
                             continue;
                         }
                         case MVM_NFA_EDGE_CODEPOINT_IM:
@@ -678,6 +676,7 @@ static MVMint64 * nqp_nfa_run(MVMThreadContext *tc, MVMNFABody *nfa, MVMString *
                             MVM_unicode_normalizer_eof(tc, &norm);
                             if (!ready)
                                 uc_arg = MVM_unicode_normalizer_get_grapheme(tc, &norm);
+                            MVM_unicode_normalizer_cleanup(tc, &norm);
 
                             MVM_unicode_normalizer_init(tc, &norm, MVM_NORMALIZE_NFD);
                             ready = MVM_unicode_normalizer_process_codepoint_to_grapheme(tc, &norm, lc_arg, &lc_arg);
@@ -688,6 +687,7 @@ static MVMint64 * nqp_nfa_run(MVMThreadContext *tc, MVMNFABody *nfa, MVMString *
                             if (((act == MVM_NFA_EDGE_CODEPOINT_IM)     && (ord == lc_arg || ord == uc_arg))
                              || ((act == MVM_NFA_EDGE_CODEPOINT_IM_NEG) && (ord != lc_arg && ord != uc_arg)))
                                 nextst[numnext++] = to;
+                            MVM_unicode_normalizer_cleanup(tc, &norm);
                             continue;
                         }
                         case MVM_NFA_EDGE_CHARRANGE_M: {
