@@ -31,20 +31,20 @@ static void deserialize_stable_size(MVMThreadContext *tc, MVMSTable *st, MVMSeri
 static void serialize_repr_data(MVMThreadContext *tc, MVMSTable *st, MVMSerializationWriter *writer) {
     MVMNativeRefREPRData *repr_data = (MVMNativeRefREPRData *)st->REPR_data;
     if (repr_data) {
-        MVM_serialization_write_varint(tc, writer, repr_data->primitive_type);
-        MVM_serialization_write_varint(tc, writer, repr_data->ref_kind);
+        MVM_serialization_write_int(tc, writer, repr_data->primitive_type);
+        MVM_serialization_write_int(tc, writer, repr_data->ref_kind);
     }
     else {
-        MVM_serialization_write_varint(tc, writer, 0);
-        MVM_serialization_write_varint(tc, writer, 0);
+        MVM_serialization_write_int(tc, writer, 0);
+        MVM_serialization_write_int(tc, writer, 0);
     }
 }
 
 /* Deserializes REPR data. */
 static void deserialize_repr_data(MVMThreadContext *tc, MVMSTable *st, MVMSerializationReader *reader) {
     MVMNativeRefREPRData *repr_data = MVM_malloc(sizeof(MVMNativeRefREPRData));
-    repr_data->primitive_type = MVM_serialization_read_varint(tc, reader);
-    repr_data->ref_kind       = MVM_serialization_read_varint(tc, reader);
+    repr_data->primitive_type = MVM_serialization_read_int(tc, reader);
+    repr_data->ref_kind       = MVM_serialization_read_int(tc, reader);
     st->REPR_data = repr_data;
 }
 
@@ -53,8 +53,8 @@ static void gc_mark(MVMThreadContext *tc, MVMSTable *st, void *data, MVMGCWorkli
     MVMNativeRefBody *ref = (MVMNativeRefBody *)data;
     MVMNativeRefREPRData *repr_data = (MVMNativeRefREPRData *)st->REPR_data;
     switch (repr_data->ref_kind) {
-        case MVM_NATIVEREF_REG_OR_LEX:
-            MVM_gc_worklist_add_frame(tc, worklist, ref->u.reg_or_lex.frame);
+        case MVM_NATIVEREF_LEX:
+            MVM_gc_worklist_add(tc, worklist, &ref->u.lex.frame);
             break;
         case MVM_NATIVEREF_ATTRIBUTE:
             MVM_gc_worklist_add(tc, worklist, &ref->u.attribute.obj);
@@ -71,14 +71,6 @@ static void gc_mark(MVMThreadContext *tc, MVMSTable *st, void *data, MVMGCWorkli
     }
 }
 
-/* Called by the VM in order to free memory associated with this object. */
-static void gc_free(MVMThreadContext *tc, MVMObject *obj) {
-    MVMNativeRef *ref = (MVMNativeRef *)obj;
-    MVMNativeRefREPRData *repr_data = (MVMNativeRefREPRData *)STABLE(obj)->REPR_data;
-    if (repr_data->ref_kind == MVM_NATIVEREF_REG_OR_LEX)
-        MVM_frame_dec_ref(tc, ref->body.u.reg_or_lex.frame);
-}
-
 /* Frees the representation data, if any. */
 static void gc_free_repr_data(MVMThreadContext *tc, MVMSTable *st) {
     MVM_free(st->REPR_data);
@@ -111,7 +103,7 @@ static void compose(MVMThreadContext *tc, MVMSTable *st, MVMObject *info_hash) {
                 MVMuint16 kind;
                 MVMString *refkind_s = MVM_repr_get_str(tc, refkind);
                 if (MVM_string_equal(tc, refkind_s, str_consts->lexical)) {
-                    kind = MVM_NATIVEREF_REG_OR_LEX;
+                    kind = MVM_NATIVEREF_LEX;
                 }
                 else if (MVM_string_equal(tc, refkind_s, str_consts->attribute)) {
                     kind = MVM_NATIVEREF_ATTRIBUTE;
@@ -156,7 +148,7 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
     if (!repr_data)
         return;
 
-    if (repr_data->ref_kind != MVM_NATIVEREF_REG_OR_LEX)
+    if (repr_data->ref_kind != MVM_NATIVEREF_LEX)
         return; /* TODO implement spesh for attribute and positional references */
 
     switch (opcode) {
@@ -169,7 +161,7 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
             ins->operands            = MVM_spesh_alloc(tc, g, 3 * sizeof(MVMSpeshOperand));
             ins->operands[0]         = target;
             ins->operands[1]         = value;
-            ins->operands[2].lit_i64 = offsetof(MVMNativeRef, body.u.reg_or_lex.var);
+            ins->operands[2].lit_i64 = offsetof(MVMNativeRef, body.u.lex.var);
             break;
         }
         case MVM_OP_decont_i: {
@@ -181,7 +173,7 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
             ins->operands            = MVM_spesh_alloc(tc, g, 3 * sizeof(MVMSpeshOperand));
             ins->operands[0]         = target;
             ins->operands[1]         = source;
-            ins->operands[2].lit_i64 = offsetof(MVMNativeRef, body.u.reg_or_lex.var);
+            ins->operands[2].lit_i64 = offsetof(MVMNativeRef, body.u.lex.var);
             break;
         }
     }
@@ -205,7 +197,7 @@ static const MVMREPROps this_repr = {
     deserialize_repr_data,
     deserialize_stable_size,
     gc_mark,
-    gc_free,
+    NULL, /* gc_free */
     NULL, /* gc_cleanup */
     NULL, /* gc_mark_repr_data */
     gc_free_repr_data,
@@ -213,7 +205,6 @@ static const MVMREPROps this_repr = {
     spesh, /* spesh */
     "NativeRef", /* name */
     MVM_REPR_ID_NativeRef,
-    1, /* refs_frames */
     NULL, /* unmanaged_size */
     NULL, /* describe_refs */
 };
@@ -235,32 +226,17 @@ void MVM_nativeref_ensure(MVMThreadContext *tc, MVMObject *type, MVMuint16 wantp
     }
 }
 
-/* Creation of native references for registers. */
-static MVMObject * reg_or_lex_ref(MVMThreadContext *tc, MVMObject *type, MVMFrame *f, MVMRegister *r, MVMuint16 reg_type) {
-    MVMNativeRef *ref = (MVMNativeRef *)MVM_gc_allocate_object(tc, STABLE(type));
-    ref->body.u.reg_or_lex.frame = MVM_frame_inc_ref(tc, f);
-    ref->body.u.reg_or_lex.var   = r;
-    ref->body.u.reg_or_lex.type  = reg_type;
+/* Creation of native references for lexicals. */
+static MVMObject * lex_ref(MVMThreadContext *tc, MVMObject *type, MVMFrame *f, MVMRegister *r, MVMuint16 reg_type) {
+    MVMNativeRef *ref;
+    MVMROOT(tc, f, {
+        ref = (MVMNativeRef *)MVM_gc_allocate_object(tc, STABLE(type));
+    });
+    MVM_ASSIGN_REF(tc, &(ref->common.header), ref->body.u.lex.frame, f);
+    ref->body.u.lex.var   = r;
+    ref->body.u.lex.type  = reg_type;
     return (MVMObject *)ref;
 }
-MVMObject * MVM_nativeref_reg_i(MVMThreadContext *tc, MVMFrame *f, MVMRegister *r) {
-    MVMObject *ref_type = MVM_hll_current(tc)->int_lex_ref;
-    if (ref_type)
-        return reg_or_lex_ref(tc, ref_type, f, r, MVM_reg_int64);
-    MVM_exception_throw_adhoc(tc, "No int lexical reference type registered for current HLL");
-}
-MVMObject * MVM_nativeref_reg_n(MVMThreadContext *tc, MVMFrame *f, MVMRegister *r) {
-    MVMObject *ref_type = MVM_hll_current(tc)->num_lex_ref;
-    if (ref_type)
-        return reg_or_lex_ref(tc, ref_type, f, r, MVM_reg_num64);
-    MVM_exception_throw_adhoc(tc, "No num lexical reference type registered for current HLL");
-}
-MVMObject * MVM_nativeref_reg_s(MVMThreadContext *tc, MVMFrame *f, MVMRegister *r) {
-    MVMObject *ref_type = MVM_hll_current(tc)->str_lex_ref;
-    if (ref_type)
-        return reg_or_lex_ref(tc, ref_type, f, r, MVM_reg_str);
-    MVM_exception_throw_adhoc(tc, "No str lexical reference type registered for current HLL");
-}
 
 /* Creation of native references for lexicals. */
 static MVMFrame * get_lexical_outer(MVMThreadContext *tc, MVMuint16 outers) {
@@ -274,7 +250,9 @@ static MVMFrame * get_lexical_outer(MVMThreadContext *tc, MVMuint16 outers) {
     return f;
 }
 MVMObject * MVM_nativeref_lex_i(MVMThreadContext *tc, MVMuint16 outers, MVMuint16 idx) {
-    MVMObject *ref_type = MVM_hll_current(tc)->int_lex_ref;
+    MVMObject *ref_type;
+    MVM_frame_force_to_heap(tc, tc->cur_frame);
+    ref_type = MVM_hll_current(tc)->int_lex_ref;
     if (ref_type) {
         MVMFrame  *f = get_lexical_outer(tc, outers);
         MVMuint16 *lexical_types = f->spesh_cand && f->spesh_cand->lexical_types
@@ -286,12 +264,14 @@ MVMObject * MVM_nativeref_lex_i(MVMThreadContext *tc, MVMuint16 outers, MVMuint1
                 type != MVM_reg_uint64 && type != MVM_reg_uint32 &&
                 type != MVM_reg_uint16 && type != MVM_reg_uint8)
             MVM_exception_throw_adhoc(tc, "getlexref_i: lexical is not an int");
-        return reg_or_lex_ref(tc, ref_type, f, &(f->env[idx]), type);
+        return lex_ref(tc, ref_type, f, &(f->env[idx]), type);
     }
     MVM_exception_throw_adhoc(tc, "No int lexical reference type registered for current HLL");
 }
 MVMObject * MVM_nativeref_lex_n(MVMThreadContext *tc, MVMuint16 outers, MVMuint16 idx) {
-    MVMObject *ref_type = MVM_hll_current(tc)->num_lex_ref;
+    MVMObject *ref_type;
+    MVM_frame_force_to_heap(tc, tc->cur_frame);
+    ref_type = MVM_hll_current(tc)->num_lex_ref;
     if (ref_type) {
         MVMFrame  *f = get_lexical_outer(tc, outers);
         MVMuint16 *lexical_types = f->spesh_cand && f->spesh_cand->lexical_types
@@ -300,12 +280,14 @@ MVMObject * MVM_nativeref_lex_n(MVMThreadContext *tc, MVMuint16 outers, MVMuint1
         MVMuint16 type = lexical_types[idx];
         if (type != MVM_reg_num64 && type != MVM_reg_num32)
             MVM_exception_throw_adhoc(tc, "getlexref_n: lexical is not a num");
-        return reg_or_lex_ref(tc, ref_type, f, &(f->env[idx]), type);
+        return lex_ref(tc, ref_type, f, &(f->env[idx]), type);
     }
     MVM_exception_throw_adhoc(tc, "No num lexical reference type registered for current HLL");
 }
 MVMObject * MVM_nativeref_lex_s(MVMThreadContext *tc, MVMuint16 outers, MVMuint16 idx) {
-    MVMObject *ref_type = MVM_hll_current(tc)->str_lex_ref;
+    MVMObject *ref_type;
+    MVM_frame_force_to_heap(tc, tc->cur_frame);
+    ref_type = MVM_hll_current(tc)->str_lex_ref;
     if (ref_type) {
         MVMFrame  *f = get_lexical_outer(tc, outers);
         MVMuint16 *lexical_types = f->spesh_cand && f->spesh_cand->lexical_types
@@ -313,13 +295,12 @@ MVMObject * MVM_nativeref_lex_s(MVMThreadContext *tc, MVMuint16 outers, MVMuint1
             : f->static_info->body.lexical_types;
         if (lexical_types[idx] != MVM_reg_str)
             MVM_exception_throw_adhoc(tc, "getlexref_s: lexical is not a str (%d, %d)", outers, idx);
-        return reg_or_lex_ref(tc, ref_type, f, &(f->env[idx]), MVM_reg_str);
+        return lex_ref(tc, ref_type, f, &(f->env[idx]), MVM_reg_str);
     }
     MVM_exception_throw_adhoc(tc, "No str lexical reference type registered for current HLL");
 }
 static MVMObject * lexref_by_name(MVMThreadContext *tc, MVMObject *type, MVMString *name, MVMuint16 kind) {
     MVMFrame *cur_frame = tc->cur_frame;
-    MVM_string_flatten(tc, name);
     while (cur_frame != NULL) {
         MVMLexicalRegistry *lexical_names = cur_frame->static_info->body.lexical_names;
         if (lexical_names) {
@@ -327,7 +308,7 @@ static MVMObject * lexref_by_name(MVMThreadContext *tc, MVMObject *type, MVMStri
             MVM_HASH_GET(tc, lexical_names, name, entry)
             if (entry) {
                 if (cur_frame->static_info->body.lexical_types[entry->value] == kind) {
-                    return reg_or_lex_ref(tc, type, cur_frame, &cur_frame->env[entry->value], kind);
+                    return lex_ref(tc, type, cur_frame, &cur_frame->env[entry->value], kind);
                 }
                 else {
                     char *c_name = MVM_string_utf8_encode_C_string(tc, name);
@@ -348,19 +329,31 @@ static MVMObject * lexref_by_name(MVMThreadContext *tc, MVMObject *type, MVMStri
     }
 }
 MVMObject * MVM_nativeref_lex_name_i(MVMThreadContext *tc, MVMString *name) {
-    MVMObject *ref_type = MVM_hll_current(tc)->int_lex_ref;
+    MVMObject *ref_type;
+    MVMROOT(tc, name, {
+        MVM_frame_force_to_heap(tc, tc->cur_frame);
+    });
+    ref_type = MVM_hll_current(tc)->int_lex_ref;
     if (ref_type)
         return lexref_by_name(tc, ref_type, name, MVM_reg_int64);
     MVM_exception_throw_adhoc(tc, "No int lexical reference type registered for current HLL");
 }
 MVMObject * MVM_nativeref_lex_name_n(MVMThreadContext *tc, MVMString *name) {
-    MVMObject *ref_type = MVM_hll_current(tc)->num_lex_ref;
+    MVMObject *ref_type;
+    MVMROOT(tc, name, {
+        MVM_frame_force_to_heap(tc, tc->cur_frame);
+    });
+    ref_type = MVM_hll_current(tc)->num_lex_ref;
     if (ref_type)
         return lexref_by_name(tc, ref_type, name, MVM_reg_num64);
     MVM_exception_throw_adhoc(tc, "No num lexical reference type registered for current HLL");
 }
 MVMObject * MVM_nativeref_lex_name_s(MVMThreadContext *tc, MVMString *name) {
-    MVMObject *ref_type = MVM_hll_current(tc)->str_lex_ref;
+    MVMObject *ref_type;
+    MVMROOT(tc, name, {
+        MVM_frame_force_to_heap(tc, tc->cur_frame);
+    });
+    ref_type = MVM_hll_current(tc)->str_lex_ref;
     if (ref_type)
         return lexref_by_name(tc, ref_type, name, MVM_reg_str);
     MVM_exception_throw_adhoc(tc, "No str lexical reference type registered for current HLL");
@@ -445,19 +438,19 @@ MVMObject * MVM_nativeref_multidim_i(MVMThreadContext *tc, MVMObject *obj, MVMOb
     MVMObject *ref_type = MVM_hll_current(tc)->int_multidim_ref;
     if (ref_type)
         return md_posref(tc, ref_type, obj, indices);
-    MVM_exception_throw_adhoc(tc, "No int positional reference type registered for current HLL");
+    MVM_exception_throw_adhoc(tc, "No int multidim positional reference type registered for current HLL");
 }
 MVMObject * MVM_nativeref_multidim_n(MVMThreadContext *tc, MVMObject *obj, MVMObject *indices) {
     MVMObject *ref_type = MVM_hll_current(tc)->num_multidim_ref;
     if (ref_type)
         return md_posref(tc, ref_type, obj, indices);
-    MVM_exception_throw_adhoc(tc, "No num positional reference type registered for current HLL");
+    MVM_exception_throw_adhoc(tc, "No num multidim positional reference type registered for current HLL");
 }
 MVMObject * MVM_nativeref_multidim_s(MVMThreadContext *tc, MVMObject *obj, MVMObject *indices) {
     MVMObject *ref_type = MVM_hll_current(tc)->str_multidim_ref;
     if (ref_type)
         return md_posref(tc, ref_type, obj, indices);
-    MVM_exception_throw_adhoc(tc, "No str positional reference type registered for current HLL");
+    MVM_exception_throw_adhoc(tc, "No str multidim positional reference type registered for current HLL");
 }
 
 /* Reference read functions. These do no checks that the reference is of the
@@ -465,31 +458,31 @@ MVMObject * MVM_nativeref_multidim_s(MVMThreadContext *tc, MVMObject *obj, MVMOb
  * they are more suited to calling from optimized code. The checking path is
  * in the native ref container implementation, in containers.c; after checks,
  * they delegate here. */
-MVMint64 MVM_nativeref_read_reg_or_lex_i(MVMThreadContext *tc, MVMObject *ref_obj) {
+MVMint64 MVM_nativeref_read_lex_i(MVMThreadContext *tc, MVMObject *ref_obj) {
     MVMNativeRef *ref = (MVMNativeRef *)ref_obj;
-    switch (ref->body.u.reg_or_lex.type) {
+    switch (ref->body.u.lex.type) {
         case MVM_reg_int8:
-            return ref->body.u.reg_or_lex.var->i8;
+            return ref->body.u.lex.var->i8;
         case MVM_reg_int16:
-            return ref->body.u.reg_or_lex.var->i16;
+            return ref->body.u.lex.var->i16;
         case MVM_reg_int32:
-            return ref->body.u.reg_or_lex.var->i32;
+            return ref->body.u.lex.var->i32;
         default:
-            return ref->body.u.reg_or_lex.var->i64;
+            return ref->body.u.lex.var->i64;
     }
 }
-MVMnum64 MVM_nativeref_read_reg_or_lex_n(MVMThreadContext *tc, MVMObject *ref_obj) {
+MVMnum64 MVM_nativeref_read_lex_n(MVMThreadContext *tc, MVMObject *ref_obj) {
     MVMNativeRef *ref = (MVMNativeRef *)ref_obj;
-    switch (ref->body.u.reg_or_lex.type) {
+    switch (ref->body.u.lex.type) {
         case MVM_reg_num32:
-            return ref->body.u.reg_or_lex.var->n32;
+            return ref->body.u.lex.var->n32;
         default:
-            return ref->body.u.reg_or_lex.var->n64;
+            return ref->body.u.lex.var->n64;
     }
 }
-MVMString * MVM_nativeref_read_reg_or_lex_s(MVMThreadContext *tc, MVMObject *ref_obj) {
+MVMString * MVM_nativeref_read_lex_s(MVMThreadContext *tc, MVMObject *ref_obj) {
     MVMNativeRef *ref = (MVMNativeRef *)ref_obj;
-    return ref->body.u.reg_or_lex.var->s;
+    return ref->body.u.lex.var->s;
 }
 MVMint64 MVM_nativeref_read_attribute_i(MVMThreadContext *tc, MVMObject *ref_obj) {
     MVMNativeRef *ref = (MVMNativeRef *)ref_obj;
@@ -532,37 +525,38 @@ MVMString * MVM_nativeref_read_multidim_s(MVMThreadContext *tc, MVMObject *ref_o
 }
 
 /* Reference write functions. Same (non-checking) rules as the reads above. */
-void MVM_nativeref_write_reg_or_lex_i(MVMThreadContext *tc, MVMObject *ref_obj, MVMint64 value) {
+void MVM_nativeref_write_lex_i(MVMThreadContext *tc, MVMObject *ref_obj, MVMint64 value) {
     MVMNativeRef *ref = (MVMNativeRef *)ref_obj;
-    switch (ref->body.u.reg_or_lex.type) {
+    switch (ref->body.u.lex.type) {
         case MVM_reg_int8:
-            ref->body.u.reg_or_lex.var->i8 = (MVMint8)value;
+            ref->body.u.lex.var->i8 = (MVMint8)value;
             break;
         case MVM_reg_int16:
-            ref->body.u.reg_or_lex.var->i16 = (MVMint16)value;
+            ref->body.u.lex.var->i16 = (MVMint16)value;
             break;
         case MVM_reg_int32:
-            ref->body.u.reg_or_lex.var->i32 = (MVMint32)value;
+            ref->body.u.lex.var->i32 = (MVMint32)value;
             break;
         default:
-            ref->body.u.reg_or_lex.var->i64 = value;
+            ref->body.u.lex.var->i64 = value;
             break;
     }
 }
-void MVM_nativeref_write_reg_or_lex_n(MVMThreadContext *tc, MVMObject *ref_obj, MVMnum64 value) {
+void MVM_nativeref_write_lex_n(MVMThreadContext *tc, MVMObject *ref_obj, MVMnum64 value) {
     MVMNativeRef *ref = (MVMNativeRef *)ref_obj;
-    switch (ref->body.u.reg_or_lex.type) {
+    switch (ref->body.u.lex.type) {
         case MVM_reg_num32:
-            ref->body.u.reg_or_lex.var->n32 = (MVMnum32)value;
+            ref->body.u.lex.var->n32 = (MVMnum32)value;
             break;
         default:
-            ref->body.u.reg_or_lex.var->n64 = value;
+            ref->body.u.lex.var->n64 = value;
             break;
     }
 }
-void MVM_nativeref_write_reg_or_lex_s(MVMThreadContext *tc, MVMObject *ref_obj, MVMString *value) {
+void MVM_nativeref_write_lex_s(MVMThreadContext *tc, MVMObject *ref_obj, MVMString *value) {
     MVMNativeRef *ref = (MVMNativeRef *)ref_obj;
-    ref->body.u.reg_or_lex.var->s = value;
+    MVM_ASSIGN_REF(tc, &(ref->body.u.lex.frame->header),
+        ref->body.u.lex.var->s, value);
 }
 void MVM_nativeref_write_attribute_i(MVMThreadContext *tc, MVMObject *ref_obj, MVMint64 value) {
     MVMNativeRef *ref = (MVMNativeRef *)ref_obj;
