@@ -349,12 +349,14 @@ static void bind_attribute(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
         case MVM_reg_obj: {
             MVMObject *value = value_reg.o;
             if (attr_st) {
-                if (attr_st == STABLE(value))
+                MVMSTable *value_st = STABLE(value);
+                if (attr_st == value_st)
                     st->REPR->copy_to(tc, attr_st, OBJECT_BODY(value), root,
                         (char *)data + repr_data->attribute_offsets[slot]);
                 else
                     MVM_exception_throw_adhoc(tc,
-                        "P6opaque: representation mismatch when storing value to attribute");
+                        "P6opaque: representation mismatch when storing value (of type %s) to attribute (of type %s)",
+                        value_st->debug_name, attr_st->debug_name);
             }
             else {
                 set_obj_at_offset(tc, root, data, repr_data->attribute_offsets[slot], value);
@@ -609,6 +611,7 @@ static void compose(MVMThreadContext *tc, MVMSTable *st, MVMObject *info_hash) {
                cur_init_slot, cur_mark_slot, cur_cleanup_slot,
                unboxed_type, i;
     MVMObject *info;
+    MVMP6opaqueREPRData *repr_data;
 
     MVMStringConsts       str_consts = tc->instance->str_consts;
     MVMString        * const str_avc = str_consts.auto_viv_container;
@@ -619,9 +622,12 @@ static void compose(MVMThreadContext *tc, MVMSTable *st, MVMObject *info_hash) {
     MVMString  * const str_attribute = str_consts.attribute;
     MVMString * const str_box_target = str_consts.box_target;
 
+    /* Check not already composed. */
+    if (st->REPR_data)
+        MVM_exception_throw_adhoc(tc, "Type %s is already composed", st->debug_name);
+
     /* Allocate the representation data. */
-    MVMP6opaqueREPRData *repr_data = MVM_malloc(sizeof(MVMP6opaqueREPRData));
-    memset(repr_data, 0, sizeof(MVMP6opaqueREPRData));
+    repr_data = (MVMP6opaqueREPRData *)MVM_calloc(1, sizeof(MVMP6opaqueREPRData));
 
     /* Find attribute information. */
     info = MVM_repr_at_key_o(tc, info_hash, str_attribute);
@@ -658,17 +664,14 @@ static void compose(MVMThreadContext *tc, MVMSTable *st, MVMObject *info_hash) {
     repr_data->num_attributes = total_attrs;
     if (total_attrs) {
         repr_data->attribute_offsets   = MVM_malloc(total_attrs * sizeof(MVMuint16));
-        repr_data->flattened_stables   = MVM_malloc(total_attrs * sizeof(MVMSTable *));
-        repr_data->auto_viv_values     = MVM_malloc(total_attrs * sizeof(MVMObject *));
+        repr_data->flattened_stables   = (MVMSTable **)MVM_calloc(total_attrs, sizeof(MVMSTable *));
+        repr_data->auto_viv_values     = (MVMObject **)MVM_calloc(total_attrs, sizeof(MVMObject *));
         repr_data->gc_obj_mark_offsets = MVM_malloc(total_attrs * sizeof(MVMuint16));
-        memset(repr_data->flattened_stables, 0, total_attrs * sizeof(MVMSTable *));
-        memset(repr_data->auto_viv_values, 0, total_attrs * sizeof(MVMObject *));
     }
-    repr_data->name_to_index_mapping = MVM_malloc((mro_count + 1) * sizeof(MVMP6opaqueNameMap));
+    repr_data->name_to_index_mapping = (MVMP6opaqueNameMap *)MVM_calloc((mro_count + 1), sizeof(MVMP6opaqueNameMap));
     repr_data->initialize_slots      = MVM_malloc((total_attrs + 1) * sizeof(MVMuint16));
     repr_data->gc_mark_slots         = MVM_malloc((total_attrs + 1) * sizeof(MVMuint16));
     repr_data->gc_cleanup_slots      = MVM_malloc((total_attrs + 1) * sizeof(MVMuint16));
-    memset(repr_data->name_to_index_mapping, 0, (mro_count + 1) * sizeof(MVMP6opaqueNameMap));
 
     /* -1 indicates no unboxing or delegate possible for a type. */
     repr_data->unbox_int_slot = -1;
