@@ -7,7 +7,7 @@
 MVMRegister * MVM_frame_initial_work(MVMThreadContext *tc, MVMuint16 *local_types,
                                      MVMuint16 num_locals) {
     MVMuint16 i;
-    MVMRegister *work_initial = MVM_calloc(sizeof(MVMRegister), num_locals);
+    MVMRegister *work_initial = MVM_calloc(num_locals, sizeof(MVMRegister));
     for (i = 0; i < num_locals; i++)
         if (local_types[i] == MVM_reg_obj)
             work_initial[i].o = tc->instance->VMNull;
@@ -27,7 +27,7 @@ static void prepare_and_verify_static_frame(MVMThreadContext *tc, MVMStaticFrame
 
     /* Take compilation unit lock, to make sure we don't race to do the
      * frame preparation/verification work. */
-    MVM_reentrantmutex_lock(tc, (MVMReentrantMutex *)cu->body.update_mutex);
+    MVM_reentrantmutex_lock(tc, (MVMReentrantMutex *)cu->body.deserialize_frame_mutex);
     if (static_frame->body.instrumentation_level == 0) {
         /* Work size is number of locals/registers plus size of the maximum
         * call site argument list. */
@@ -64,7 +64,7 @@ static void prepare_and_verify_static_frame(MVMThreadContext *tc, MVMStaticFrame
     }
 
     /* Unlock, now we're finished. */
-    MVM_reentrantmutex_unlock(tc, (MVMReentrantMutex *)cu->body.update_mutex);
+    MVM_reentrantmutex_unlock(tc, (MVMReentrantMutex *)cu->body.deserialize_frame_mutex);
 }
 
 /* When we don't match the current instrumentation level, we hit this. It may
@@ -88,7 +88,9 @@ static void instrumentation_level_barrier(MVMThreadContext *tc, MVMStaticFrame *
         MVM_profile_ensure_uninstrumented(tc, static_frame);
 }
 
-/* Destroys a frame. */
+/* Called when the GC destroys a frame. Since the frame may have been alive as
+ * part of a continuation that was taken but never invoked, we should check
+ * things normally cleaned up on return don't need cleaning up also. */
 void MVM_frame_destroy(MVMThreadContext *tc, MVMFrame *frame) {
     if (frame->work) {
         MVM_args_proc_cleanup(tc, &frame->params);
@@ -117,9 +119,6 @@ static MVMFrame * create_context_only(MVMThreadContext *tc, MVMStaticFrame *stat
         frame = MVM_gc_allocate_frame(tc);
     });
     });
-    
-    /* Copy thread context into the frame. */
-    frame->tc = tc;
 
     /* Set static frame. */
     MVM_ASSIGN_REF(tc, &(frame->header), frame->static_info, static_frame);
@@ -414,13 +413,7 @@ void MVM_frame_invoke(MVMThreadContext *tc, MVMStaticFrame *static_frame,
     if (spesh_cand >= 0 && spesh_cand < static_frame->body.num_spesh_candidates) {
         MVMSpeshCandidate *chosen_cand = &static_frame->body.spesh_candidates[spesh_cand];
         if (!chosen_cand->sg) {
-            MVMROOT(tc, static_frame, {
-            MVMROOT(tc, code_ref, {
-            MVMROOT(tc, outer, {
-                frame = allocate_frame(tc, static_frame, chosen_cand);
-            });
-            });
-            });
+            frame = allocate_frame(tc, static_frame, chosen_cand);
             frame->effective_bytecode    = chosen_cand->bytecode;
             frame->effective_handlers    = chosen_cand->handlers;
             frame->effective_spesh_slots = chosen_cand->spesh_slots;
@@ -520,13 +513,7 @@ void MVM_frame_invoke(MVMThreadContext *tc, MVMStaticFrame *static_frame,
                 if (!chosen_cand->osr_logging && cur_idx < MVM_SPESH_LOG_RUNS) {
                     if (MVM_cas(&(chosen_cand->log_enter_idx), cur_idx, cur_idx + 1) == cur_idx) {
                         /* We get to log. */
-                        MVMROOT(tc, static_frame, {
-                        MVMROOT(tc, code_ref, {
-                        MVMROOT(tc, outer, {
-                            frame = allocate_frame(tc, static_frame, chosen_cand);
-                        });
-                        });
-                        });
+                        frame = allocate_frame(tc, static_frame, chosen_cand);
                         frame->effective_bytecode    = chosen_cand->bytecode;
                         frame->effective_handlers    = chosen_cand->handlers;
                         frame->effective_spesh_slots = chosen_cand->spesh_slots;
@@ -539,13 +526,7 @@ void MVM_frame_invoke(MVMThreadContext *tc, MVMStaticFrame *static_frame,
             }
             else {
                 /* In the post-specialize phase; can safely used the code. */
-                MVMROOT(tc, static_frame, {
-                MVMROOT(tc, code_ref, {
-                MVMROOT(tc, outer, {
-                    frame = allocate_frame(tc, static_frame, chosen_cand);
-                });
-                });
-                });
+                frame = allocate_frame(tc, static_frame, chosen_cand);
                 if (chosen_cand->jitcode) {
                     frame->effective_bytecode = chosen_cand->jitcode->bytecode;
                     frame->jit_entry_label    = chosen_cand->jitcode->labels[0];
@@ -562,20 +543,11 @@ void MVM_frame_invoke(MVMThreadContext *tc, MVMStaticFrame *static_frame,
         }
     }
     if (!found_spesh) {
-        MVMROOT(tc, static_frame, {
-        MVMROOT(tc, code_ref, {
-        MVMROOT(tc, outer, {
-            frame = allocate_frame(tc, static_frame, NULL);
-        });
-        });
-        });
+        frame = allocate_frame(tc, static_frame, NULL);
         frame->effective_bytecode = static_frame->body.bytecode;
         frame->effective_handlers = static_frame->body.handlers;
     }
 
-    /* Copy thread context into the frame. */
-    frame->tc = tc;
-
     /* Set static frame. */
     frame->static_info = static_frame;
 
@@ -632,8 +604,7 @@ void MVM_frame_invoke(MVMThreadContext *tc, MVMStaticFrame *static_frame,
                         }
                         else {
                             /* Allocate storage for state vars. */
-                            state = MVM_malloc(frame->static_info->body.env_size);
-                            memset(state, 0, frame->static_info->body.env_size);
+                            state = (MVMRegister *)MVM_calloc(1, frame->static_info->body.env_size);
                             ((MVMCode *)frame->code_ref)->body.state_vars = state;
                             state_act = 1;
 
@@ -766,7 +737,6 @@ MVMFrame * MVM_frame_create_for_deopt(MVMThreadContext *tc, MVMStaticFrame *stat
     });
     frame->effective_bytecode       = static_frame->body.bytecode;
     frame->effective_handlers       = static_frame->body.handlers;
-    frame->tc                       = tc;
     MVM_ASSIGN_REF(tc, &(frame->header), frame->static_info, static_frame);
     MVM_ASSIGN_REF(tc, &(frame->header), frame->code_ref, code_ref);
     MVM_ASSIGN_REF(tc, &(frame->header), frame->outer, code_ref->body.outer);
@@ -797,23 +767,15 @@ static MVMuint64 remove_one_frame(MVMThreadContext *tc, MVMuint8 unwind) {
         }
     }
 
-    /* Some cleanup we only need do if we're not a frame involved in a
-     * continuation (otherwise we need to allow for multi-shot
-     * re-invocation). */
-    if (!returner->in_continuation) {
-        /* Arguments buffer no longer in use (saves GC visiting it). */
-        returner->cur_args_callsite = NULL;
+    /* Clear up any continuation tags. */
+    if (returner->continuation_tags)
+        MVM_continuation_free_tags(tc, returner);
 
-        /* Clear up argument processing leftovers, if any. */
-        if (returner->work)
-            MVM_args_proc_cleanup_for_cache(tc, &returner->params);
-
-        /* Clear up any continuation tags. */
-        if (returner->continuation_tags)
-            MVM_continuation_free_tags(tc, returner);
-
-        /* Signal to the GC to ignore ->work */
-        returner->tc = NULL;
+    /* Clean up frame working space. */
+    if (returner->work) {
+        MVM_args_proc_cleanup(tc, &returner->params);
+        MVM_fixed_size_free(tc, tc->instance->fsa, returner->allocd_work,
+            returner->work);
     }
 
     /* If it's a call stack frame, remove it from the stack. */
@@ -822,7 +784,17 @@ static MVMuint64 remove_one_frame(MVMThreadContext *tc, MVMuint8 unwind) {
         stack->alloc = (char *)returner;
         if ((char *)stack->alloc - sizeof(MVMCallStackRegion) == (char *)stack)
             MVM_callstack_region_prev(tc);
-        MVM_frame_destroy(tc, returner);
+        if (returner->env)
+            MVM_fixed_size_free(tc, tc->instance->fsa, returner->allocd_env, returner->env);
+    }
+
+    /* Otherwise, NULL  out ->work, to indicate the frame is no longer in
+     * dynamic scope. This is used by the GC to avoid marking stuff (this is
+     * needed for safety as otherwise we'd read freed memory), as well as by
+     * exceptions to ensure the target of an exception throw is indeed still
+     * in dynamic scope. */
+    else {
+        returner->work = NULL;
     }
 
     /* Switch back to the caller frame if there is one. */
@@ -883,21 +855,25 @@ MVMuint64 MVM_frame_try_return(MVMThreadContext *tc) {
         if (tc->cur_frame == tc->thread_entry_frame)
             MVM_exception_throw_adhoc(tc, "Thread entry point frame cannot have an exit handler");
 
-        switch (caller->return_type) {
-            case MVM_RETURN_OBJ:
-                result = caller->return_value->o;
-                break;
-            case MVM_RETURN_INT:
-                result = MVM_repr_box_int(tc, hll->int_box_type, caller->return_value->i64);
-                break;
-            case MVM_RETURN_NUM:
-                result = MVM_repr_box_num(tc, hll->num_box_type, caller->return_value->n64);
-                break;
-            case MVM_RETURN_STR:
-                result = MVM_repr_box_str(tc, hll->str_box_type, caller->return_value->s);
-                break;
-            default:
-                result = NULL;
+        if (caller->return_type == MVM_RETURN_OBJ) {
+            result = caller->return_value->o;
+        }
+        else {
+            MVMROOT(tc, cur_frame, {
+                switch (caller->return_type) {
+                    case MVM_RETURN_INT:
+                        result = MVM_repr_box_int(tc, hll->int_box_type, caller->return_value->i64);
+                        break;
+                    case MVM_RETURN_NUM:
+                        result = MVM_repr_box_num(tc, hll->num_box_type, caller->return_value->n64);
+                        break;
+                    case MVM_RETURN_STR:
+                        result = MVM_repr_box_str(tc, hll->str_box_type, caller->return_value->s);
+                        break;
+                    default:
+                        result = NULL;
+                }
+            });
         }
 
         handler = MVM_frame_find_invokee(tc, hll->exit_handler, NULL);
@@ -1068,7 +1044,9 @@ void MVM_frame_capture_inner(MVMThreadContext *tc, MVMObject *code) {
     MVMFrame *outer;
     MVMROOT(tc, code, {
         MVMStaticFrame *sf_outer = code_obj->body.sf->body.outer;
-        outer = create_context_only(tc, sf_outer, (MVMObject *)sf_outer->body.static_code, 1);
+        MVMROOT(tc, sf_outer, {
+            outer = create_context_only(tc, sf_outer, (MVMObject *)sf_outer->body.static_code, 1);
+        });
         MVMROOT(tc, outer, {
             MVMFrame *outer_outer = autoclose(tc, sf_outer->body.outer);
             MVM_ASSIGN_REF(tc, &(outer->header), outer->outer, outer_outer);
@@ -1792,32 +1770,3 @@ MVMObject * MVM_frame_context_wrapper(MVMThreadContext *tc, MVMFrame *f) {
     });
     return ctx;
 }
-
-/* Creates a shallow clone of a frame. Used by continuations. We rely on this
- * not allocating with the GC; update continuation clone code if it comes to
-  * do so. */
-MVMFrame * MVM_frame_clone(MVMThreadContext *tc, MVMFrame *f) {
-    /* First, just grab a copy of everything. */
-    MVMFrame *clone;
-    MVMROOT(tc, f, {
-        clone = MVM_gc_allocate_frame(tc);
-    });
-    memcpy(
-        (char *)clone + sizeof(MVMCollectable),
-        (char *)f + sizeof(MVMCollectable),
-        sizeof(MVMFrame) - sizeof(MVMCollectable));
-
-    /* Need fresh env and work. */
-    if (f->static_info->body.env_size) {
-        clone->env = MVM_fixed_size_alloc(tc, tc->instance->fsa, f->static_info->body.env_size);
-        clone->allocd_env = f->static_info->body.env_size;
-        memcpy(clone->env, f->env, f->static_info->body.env_size);
-    }
-    if (f->static_info->body.work_size) {
-        clone->work = MVM_malloc(f->static_info->body.work_size);
-        memcpy(clone->work, f->work, f->static_info->body.work_size);
-        clone->args = clone->work + f->static_info->body.num_locals;
-    }
-
-    return clone;
-}
