@@ -271,6 +271,20 @@ void MVM_gc_mark_thread_unblocked(MVMThreadContext *tc) {
     }
 }
 
+/* Checks if a thread has marked itself as blocked. Considers that the GC may
+ * have stolen its work and marked it as such also. So what this really
+ * answers is, "did this thread mark itself blocked, and since then not mark
+ * itself unblocked", which is useful if you need to conditionally unblock
+ * or re-block. If the status changes from blocked to stolen or stolen to
+ * blocked between checking this and calling unblock, it's safe anyway since
+ * these cases are handled in MVM_gc_mark_thread_unblocked. Note that this
+ * relies on a thread itself only ever calling block/unblock. */
+MVMint32 MVM_gc_is_thread_blocked(MVMThreadContext *tc) {
+    AO_t gc_status = MVM_load(&(tc->gc_status));
+    return gc_status == MVMGCStatus_UNABLE ||
+           gc_status == MVMGCStatus_STOLEN;
+}
+
 static MVMint32 is_full_collection(MVMThreadContext *tc) {
     MVMuint64 percent_growth;
     size_t rss, promoted;
@@ -280,6 +294,11 @@ static MVMint32 is_full_collection(MVMThreadContext *tc) {
     if (promoted < MVM_GC_GEN2_THRESHOLD_MINIMUM)
         return 0;
 
+    /* If we're heap profiling then don't consider the resident set size, as
+     * it will be hugely distorted by the profile data we record. */
+    if (MVM_profile_heap_profiling(tc))
+        return 1;
+
     /* Otherwise, consider percentage of resident set size. */
     if (uv_resident_set_memory(&rss) < 0 || rss == 0)
         rss = 50 * 1024 * 1024;
@@ -291,6 +310,11 @@ static void run_gc(MVMThreadContext *tc, MVMuint8 what_to_do) {
     MVMuint8   gen;
     MVMuint32  i, n;
 
+#if MVM_GC_DEBUG
+    if (tc->in_spesh)
+        MVM_panic(1, "Must not GC when in the specializer/JIT\n");
+#endif
+
     /* Decide nursery or full collection. */
     gen = tc->instance->gc_full_collect ? MVMGCGenerations_Both : MVMGCGenerations_Nursery;
 
@@ -495,12 +519,6 @@ void MVM_gc_enter_from_interrupt(MVMThreadContext *tc) {
 void MVM_gc_global_destruction(MVMThreadContext *tc) {
     char *nursery_tmp;
 
-    /* Must wait until we're the only thread... */
-    while (tc->instance->num_user_threads) {
-        GC_SYNC_POINT(tc);
-        MVM_platform_thread_yield();
-    }
-
     /* Fake a nursery collection run by swapping the semi-
      * space nurseries. */
     nursery_tmp = tc->nursery_fromspace;
