@@ -1,8 +1,5 @@
 #include "moar.h"
 
-/* Number of bytes we accept per read. */
-#define CHUNK_SIZE 65536
-
 /* Data that we keep for an asynchronous socket handle. */
 typedef struct {
     /* The libuv handle to the socket. */
@@ -29,6 +26,11 @@ static void on_alloc(uv_handle_t *handle, size_t suggested_size, uv_buf_t *buf)
     buf->len    = size;
 }
 
+/* Callback used to simply free memory on close. */
+static void free_on_close_cb(uv_handle_t *handle) {
+    MVM_free(handle);
+}
+
 /* Read handler. */
 static void on_read(uv_stream_t *handle, ssize_t nread, const uv_buf_t *buf) {
     ReadInfo         *ri  = (ReadInfo *)handle->data;
@@ -37,7 +39,7 @@ static void on_read(uv_stream_t *handle, ssize_t nread, const uv_buf_t *buf) {
     MVMAsyncTask     *t   = (MVMAsyncTask *)MVM_repr_at_pos_o(tc,
         tc->instance->event_loop_active, ri->work_idx);
     MVM_repr_push_o(tc, arr, t->body.schedulee);
-    if (nread > 0) {
+    if (nread >= 0) {
         MVMROOT(tc, t, {
         MVMROOT(tc, arr, {
             /* Push the sequence number. */
@@ -58,7 +60,7 @@ static void on_read(uv_stream_t *handle, ssize_t nread, const uv_buf_t *buf) {
                 MVMArray *res_buf      = (MVMArray *)MVM_repr_alloc_init(tc, ri->buf_type);
                 res_buf->body.slots.i8 = (MVMint8 *)buf->base;
                 res_buf->body.start    = 0;
-                res_buf->body.ssize    = nread;
+                res_buf->body.ssize    = buf->len;
                 res_buf->body.elems    = nread;
                 MVM_repr_push_o(tc, arr, (MVMObject *)res_buf);
             }
@@ -147,7 +149,7 @@ static void read_gc_free(MVMThreadContext *tc, MVMObject *t, void *data) {
     if (data) {
         ReadInfo *ri = (ReadInfo *)data;
         if (ri->ds)
-            MVM_string_decodestream_destory(tc, ri->ds);
+            MVM_string_decodestream_destroy(tc, ri->ds);
         MVM_free(data);
     }
 }
@@ -190,7 +192,9 @@ static MVMAsyncTask * read_chars(MVMThreadContext *tc, MVMOSHandle *h, MVMObject
     task->body.data = ri;
 
     /* Hand the task off to the event loop. */
-    MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    MVMROOT(tc, task, {
+        MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    });
 
     return task;
 }
@@ -235,7 +239,9 @@ static MVMAsyncTask * read_bytes(MVMThreadContext *tc, MVMOSHandle *h, MVMObject
     task->body.data = ri;
 
     /* Hand the task off to the event loop. */
-    MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    MVMROOT(tc, task, {
+        MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    });
 
     return task;
 }
@@ -398,7 +404,9 @@ static MVMAsyncTask * write_str(MVMThreadContext *tc, MVMOSHandle *h, MVMObject
     task->body.data = wi;
 
     /* Hand the task off to the event loop. */
-    MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    MVMROOT(tc, task, {
+        MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    });
 
     return task;
 }
@@ -440,22 +448,21 @@ static MVMAsyncTask * write_bytes(MVMThreadContext *tc, MVMOSHandle *h, MVMObjec
     task->body.data = wi;
 
     /* Hand the task off to the event loop. */
-    MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    MVMROOT(tc, task, {
+        MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    });
 
     return task;
 }
 
 /* Does an asynchronous close (since it must run on the event loop). */
-static void close_cb(uv_handle_t *handle) {
-    MVM_free(handle);
-}
 static void close_perform(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_task, void *data) {
     uv_handle_t *handle = (uv_handle_t *)data;
 
     if (uv_is_closing(handle))
         MVM_exception_throw_adhoc(tc, "cannot close a closed socket");
 
-    uv_close(handle, close_cb);
+    uv_close(handle, free_on_close_cb);
 }
 
 /* Operations table for async close task. */
@@ -484,7 +491,7 @@ static MVMint64 close_socket(MVMThreadContext *tc, MVMOSHandle *h) {
 static void gc_free(MVMThreadContext *tc, MVMObject *h, void *d) {
     MVMIOAsyncSocketData *data = (MVMIOAsyncSocketData *)d;
     if (data->ds) {
-        MVM_string_decodestream_destory(tc, data->ds);
+        MVM_string_decodestream_destroy(tc, data->ds);
         data->ds = NULL;
     }
 }
@@ -590,10 +597,10 @@ static void connect_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *asyn
         });
 
         /* Cleanup handles. */
-        MVM_free(ci->socket);
-        ci->socket = NULL;
         MVM_free(ci->connect);
         ci->connect = NULL;
+        uv_close((uv_handle_t *)ci->socket, free_on_close_cb);
+        ci->socket = NULL;
     }
 }
 
@@ -647,7 +654,9 @@ MVMObject * MVM_io_socket_connect_async(MVMThreadContext *tc, MVMObject *queue,
     task->body.data = ci;
 
     /* Hand the task off to the event loop. */
-    MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    MVMROOT(tc, task, {
+        MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    });
 
     return (MVMObject *)task;
 }
@@ -736,7 +745,7 @@ static void listen_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async
             });
             MVM_repr_push_o(tc, t->body.queue, arr);
         });
-        MVM_free(li->socket);
+        uv_close((uv_handle_t *)li->socket, free_on_close_cb);
         li->socket = NULL;
         return;
     }
@@ -806,7 +815,9 @@ MVMObject * MVM_io_socket_listen_async(MVMThreadContext *tc, MVMObject *queue,
     task->body.data = li;
 
     /* Hand the task off to the event loop. */
-    MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    MVMROOT(tc, task, {
+        MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    });
 
     return (MVMObject *)task;
 }
