@@ -29,6 +29,11 @@ static void on_alloc(uv_handle_t *handle, size_t suggested_size, uv_buf_t *buf)
     buf->len    = size;
 }
 
+/* Callback used to simply free memory on close. */
+static void free_on_close_cb(uv_handle_t *handle) {
+    MVM_free(handle);
+}
+
 /* Read handler. */
 static void on_read(uv_udp_t *handle, ssize_t nread, const uv_buf_t *buf, const struct sockaddr *addr, unsigned flags) {
     ReadInfo         *ri  = (ReadInfo *)handle->data;
@@ -58,7 +63,7 @@ static void on_read(uv_udp_t *handle, ssize_t nread, const uv_buf_t *buf, const
                 MVMArray *res_buf      = (MVMArray *)MVM_repr_alloc_init(tc, ri->buf_type);
                 res_buf->body.slots.i8 = (MVMint8 *)buf->base;
                 res_buf->body.start    = 0;
-                res_buf->body.ssize    = nread;
+                res_buf->body.ssize    = buf->len;
                 res_buf->body.elems    = nread;
                 MVM_repr_push_o(tc, arr, (MVMObject *)res_buf);
             }
@@ -147,7 +152,7 @@ static void read_gc_free(MVMThreadContext *tc, MVMObject *t, void *data) {
     if (data) {
         ReadInfo *ri = (ReadInfo *)data;
         if (ri->ds)
-            MVM_string_decodestream_destory(tc, ri->ds);
+            MVM_string_decodestream_destroy(tc, ri->ds);
         MVM_free(data);
     }
 }
@@ -190,7 +195,9 @@ static MVMAsyncTask * read_chars(MVMThreadContext *tc, MVMOSHandle *h, MVMObject
     task->body.data = ri;
 
     /* Hand the task off to the event loop. */
-    MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    MVMROOT(tc, task, {
+        MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    });
 
     return task;
 }
@@ -235,7 +242,9 @@ static MVMAsyncTask * read_bytes(MVMThreadContext *tc, MVMOSHandle *h, MVMObject
     task->body.data = ri;
 
     /* Hand the task off to the event loop. */
-    MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    MVMROOT(tc, task, {
+        MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    });
 
     return task;
 }
@@ -409,7 +418,9 @@ static MVMAsyncTask * write_str_to(MVMThreadContext *tc, MVMOSHandle *h, MVMObje
     task->body.data = wi;
 
     /* Hand the task off to the event loop. */
-    MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    MVMROOT(tc, task, {
+        MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    });
 
     return task;
 }
@@ -457,22 +468,21 @@ static MVMAsyncTask * write_bytes_to(MVMThreadContext *tc, MVMOSHandle *h, MVMOb
     task->body.data = wi;
 
     /* Hand the task off to the event loop. */
-    MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    MVMROOT(tc, task, {
+        MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    });
 
     return task;
 }
 
 /* Does an asynchronous close (since it must run on the event loop). */
-static void close_cb(uv_handle_t *handle) {
-    MVM_free(handle);
-}
 static void close_perform(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_task, void *data) {
     uv_handle_t *handle = (uv_handle_t *)data;
 
     if (uv_is_closing(handle))
         MVM_exception_throw_adhoc(tc, "cannot close a closed socket");
 
-    uv_close(handle, close_cb);
+    uv_close(handle, free_on_close_cb);
 }
 
 /* Operations table for async close task. */
@@ -501,7 +511,7 @@ static MVMint64 close_socket(MVMThreadContext *tc, MVMOSHandle *h) {
 static void gc_free(MVMThreadContext *tc, MVMObject *h, void *d) {
     MVMIOAsyncUDPSocketData *data = (MVMIOAsyncUDPSocketData *)d;
     if (data->ds) {
-        MVM_string_decodestream_destory(tc, data->ds);
+        MVM_string_decodestream_destroy(tc, data->ds);
         data->ds = NULL;
     }
 }
@@ -581,7 +591,7 @@ static void setup_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
             });
             });
             MVM_repr_push_o(tc, t->body.queue, arr);
-            MVM_free(udp_handle);
+            uv_close((uv_handle_t *)udp_handle, free_on_close_cb);
         });
     }
 }
@@ -640,7 +650,9 @@ MVMObject * MVM_io_socket_udp_async(MVMThreadContext *tc, MVMObject *queue,
     task->body.data = ssi;
 
     /* Hand the task off to the event loop. */
-    MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    MVMROOT(tc, task, {
+        MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    });
 
     return (MVMObject *)task;
 }
