@@ -529,7 +529,9 @@ static MVMAsyncTask * write_str(MVMThreadContext *tc, MVMOSHandle *h, MVMObject
     task->body.data = wi;
 
     /* Hand the task off to the event loop. */
-    MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    MVMROOT(tc, task, {
+        MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    });
 
     return task;
 }
@@ -571,7 +573,9 @@ static MVMAsyncTask * write_bytes(MVMThreadContext *tc, MVMOSHandle *h, MVMObjec
     task->body.data = wi;
 
     /* Hand the task off to the event loop. */
-    MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    MVMROOT(tc, task, {
+        MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    });
 
     return task;
 }
@@ -734,11 +738,14 @@ static void on_alloc(uv_handle_t *handle, size_t suggested_size, uv_buf_t *buf)
 static void async_read(uv_stream_t *handle, ssize_t nread, const uv_buf_t *buf, SpawnInfo *si,
                        MVMObject *callback, MVMDecodeStream *ds, MVMuint32 seq_number) {
     MVMThreadContext *tc  = si->tc;
-    MVMObject        *arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
-    MVMAsyncTask     *t   = (MVMAsyncTask *)MVM_repr_at_pos_o(tc,
-        tc->instance->event_loop_active, si->work_idx);
+    MVMObject *arr;
+    MVMAsyncTask *t;
+    MVMROOT(tc, callback, {
+        arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
+        t = (MVMAsyncTask *)MVM_repr_at_pos_o(tc, tc->instance->event_loop_active, si->work_idx);
+    });
     MVM_repr_push_o(tc, arr, callback);
-    if (nread > 0) {
+    if (nread >= 0) {
         MVMROOT(tc, t, {
         MVMROOT(tc, arr, {
             /* Push the sequence number. */
@@ -761,7 +768,7 @@ static void async_read(uv_stream_t *handle, ssize_t nread, const uv_buf_t *buf,
                 MVMArray  *res_buf     = (MVMArray *)MVM_repr_alloc_init(tc, buf_type);
                 res_buf->body.slots.i8 = (MVMint8 *)buf->base;
                 res_buf->body.start    = 0;
-                res_buf->body.ssize    = nread;
+                res_buf->body.ssize    = buf->len;
                 res_buf->body.elems    = nread;
                 MVM_repr_push_o(tc, arr, (MVMObject *)res_buf);
             }
@@ -1017,11 +1024,11 @@ static void spawn_gc_free(MVMThreadContext *tc, MVMObject *t, void *data) {
             si->args = NULL;
         }
         if (si->ds_stdout) {
-            MVM_string_decodestream_destory(tc, si->ds_stdout);
+            MVM_string_decodestream_destroy(tc, si->ds_stdout);
             si->ds_stdout = NULL;
         }
         if (si->ds_stderr) {
-            MVM_string_decodestream_destory(tc, si->ds_stderr);
+            MVM_string_decodestream_destroy(tc, si->ds_stderr);
             si->ds_stderr = NULL;
         }
         MVM_free(si);
@@ -1105,7 +1112,9 @@ MVMObject * MVM_proc_spawn_async(MVMThreadContext *tc, MVMObject *queue, MVMObje
     });
 
     /* Hand the task off to the event loop. */
-    MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    MVMROOT(tc, handle, {
+        MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    });
 
     return (MVMObject *)handle;
 }
