@@ -46,7 +46,7 @@ static MVMint64 closefh(MVMThreadContext *tc, MVMOSHandle *h) {
     MVMIOFileData *data = (MVMIOFileData *)h->body.data;
     uv_fs_t req;
     if (data->ds) {
-        MVM_string_decodestream_destory(tc, data->ds);
+        MVM_string_decodestream_destroy(tc, data->ds);
         data->ds = NULL;
     }
     if (uv_fs_close(tc->loop, &req, data->fd, NULL) < 0) {
@@ -84,7 +84,7 @@ static void seek(MVMThreadContext *tc, MVMOSHandle *h, MVMint64 offset, MVMint64
 
     if (data->ds) {
         /* We'll start over from a new position. */
-        MVM_string_decodestream_destory(tc, data->ds);
+        MVM_string_decodestream_destroy(tc, data->ds);
         data->ds = NULL;
     }
 
@@ -99,7 +99,15 @@ static void seek(MVMThreadContext *tc, MVMOSHandle *h, MVMint64 offset, MVMint64
 /* Get curernt position in the file. */
 static MVMint64 mvm_tell(MVMThreadContext *tc, MVMOSHandle *h) {
     MVMIOFileData *data = (MVMIOFileData *)h->body.data;
-    return data->ds ? MVM_string_decodestream_tell_bytes(tc, data->ds) : 0;
+    MVMint64 r;
+
+    if (data->ds)
+        return MVM_string_decodestream_tell_bytes(tc, data->ds);
+
+    if ((r = MVM_platform_lseek(data->fd, 0, SEEK_CUR)) == -1)
+        MVM_exception_throw_adhoc(tc, "Failed to tell in filehandle: %d", errno);
+
+    return r;
 }
 
 /* Set the line separator. */
@@ -114,12 +122,15 @@ static MVMint32 read_to_buffer(MVMThreadContext *tc, MVMIOFileData *data, MVMint
     uv_buf_t read_buf = uv_buf_init(buf, bytes);
     uv_fs_t req;
     MVMint32 read;
+    MVM_gc_mark_thread_blocked(tc);
     if ((read = uv_fs_read(tc->loop, &req, data->fd, &read_buf, 1, -1, NULL)) < 0) {
         MVM_free(buf);
+        MVM_gc_mark_thread_unblocked(tc);
         MVM_exception_throw_adhoc(tc, "Reading from filehandle failed: %s",
             uv_strerror(req.result));
     }
     MVM_string_decodestream_add_bytes(tc, data->ds, buf, read);
+    MVM_gc_mark_thread_unblocked(tc);
     return read;
 }
 
@@ -220,7 +231,10 @@ static MVMint64 mvm_eof(MVMThreadContext *tc, MVMOSHandle *h) {
     }
     if ((seek_pos = MVM_platform_lseek(data->fd, 0, SEEK_CUR)) == -1)
         MVM_exception_throw_adhoc(tc, "Failed to seek in filehandle: %d", errno);
-    return req.statbuf.st_size == seek_pos;
+    /* Comparison with seek_pos for some special files, like those in /proc,
+     * which file size is 0 can be false. In that case, we fall back to check
+     * file size to detect EOF. */
+    return req.statbuf.st_size == seek_pos || req.statbuf.st_size == 0;
 }
 
 /* Writes the specified string to the file handle, maybe with a newline. */
@@ -389,7 +403,7 @@ static void gc_free(MVMThreadContext *tc, MVMObject *h, void *d) {
     MVMIOFileData *data = (MVMIOFileData *)d;
     if (data) {
         if (data->ds)
-            MVM_string_decodestream_destory(tc, data->ds);
+            MVM_string_decodestream_destroy(tc, data->ds);
         MVM_string_decode_stream_sep_destroy(tc, &(data->sep_spec));
         if (data->filename)
             MVM_free(data->filename);
@@ -456,37 +470,69 @@ static int resolve_open_mode(int *flag, const char *cp) {
 
 /* Opens a file, returning a synchronous file handle. */
 MVMObject * MVM_file_open_fh(MVMThreadContext *tc, MVMString *filename, MVMString *mode) {
-    char          * const fname  = MVM_string_utf8_c8_encode_C_string(tc, filename);
-    char          * const fmode  = MVM_string_utf8_encode_C_string(tc, mode);
-    MVMOSHandle   * const result = (MVMOSHandle *)MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTIO);
-    MVMIOFileData * const data   = MVM_calloc(1, sizeof(MVMIOFileData));
+    char * const fname = MVM_string_utf8_c8_encode_C_string(tc, filename);
     uv_fs_t req;
     uv_file fd;
+    int flag;
 
     /* Resolve mode description to flags. */
-    int flag;
-    if (!resolve_open_mode(&flag, fmode)) {
-        char *waste[] = { fmode, NULL };
-        MVM_free(fname);
-        MVM_exception_throw_adhoc_free(tc, waste, "Invalid open mode: %s", fmode);
+    {
+        char * const fmode  = MVM_string_utf8_encode_C_string(tc, mode);
+
+        if (!resolve_open_mode(&flag, fmode)) {
+            char *waste[] = { fname, fmode, NULL };
+            MVM_exception_throw_adhoc_free(tc, waste, "Invalid open mode for file %s: %s", fname, fmode);
+        }
+        MVM_free(fmode);
     }
-    MVM_free(fmode);
 
     /* Try to open the file. */
     if ((fd = uv_fs_open(tc->loop, &req, (const char *)fname, flag, DEFAULT_MODE, NULL)) < 0) {
         char *waste[] = { fname, NULL };
-        MVM_exception_throw_adhoc_free(tc, waste, "Failed to open file %s: %s", fname, uv_strerror(req.result));
+        const char *err = uv_strerror(req.result);
+
+        uv_fs_req_cleanup(&req);
+        MVM_exception_throw_adhoc_free(tc, waste, "Failed to open file %s: %s", fname, err);
     }
+    uv_fs_req_cleanup(&req);
+
+    /*  Check that we didn't open a directory by accident.
+        If fstat fails, just move on: Most of the documented error cases should
+        already have triggered when opening the file, and we can't do anything
+        about the others; a failure also does not necessarily imply that the
+        file descriptor cannot be used for reading/writing. */
+    if (uv_fs_fstat(tc->loop, &req, fd, NULL) == 0 && (req.statbuf.st_mode & S_IFMT) == S_IFDIR) {
+        char *waste[] = { fname, NULL };
+
+        uv_fs_req_cleanup(&req);
+
+        if (uv_fs_close(tc->loop, &req, fd, NULL) < 0) {
+            const char *err = uv_strerror(req.result);
+
+            uv_fs_req_cleanup(&req);
+            MVM_exception_throw_adhoc_free(tc, waste, "Tried to open directory %s, which we failed to close: %s",
+                fname, err);
+        }
+        uv_fs_req_cleanup(&req);
+
+        MVM_exception_throw_adhoc_free(tc, waste, "Tried to open directory %s", fname);
+    }
+    uv_fs_req_cleanup(&req);
 
     /* Set up handle. */
-    data->fd          = fd;
-    data->filename    = fname;
-    data->encoding    = MVM_encoding_type_utf8;
-    MVM_string_decode_stream_sep_default(tc, &(data->sep_spec));
-    result->body.ops  = &op_table;
-    result->body.data = data;
+    {
+        MVMIOFileData * const data   = MVM_calloc(1, sizeof(MVMIOFileData));
+        MVMOSHandle   * const result = (MVMOSHandle *)MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTIO);
 
-    return (MVMObject *)result;
+        data->fd          = fd;
+        data->filename    = fname;
+        data->encoding    = MVM_encoding_type_utf8;
+        MVM_string_decode_stream_sep_default(tc, &(data->sep_spec));
+        result->body.ops  = &op_table;
+        result->body.data = data;
+
+        return (MVMObject *)result;
+    }
 }
 
 /* Opens a file, returning a synchronous file handle. */
