@@ -25,7 +25,7 @@ static MVMint64 do_close(MVMThreadContext *tc, MVMIOSyncSocketData *data) {
          data->ss.handle = NULL;
     }
     if (data->ss.ds) {
-        MVM_string_decodestream_destory(tc, data->ss.ds);
+        MVM_string_decodestream_destroy(tc, data->ss.ds);
         data->ss.ds = NULL;
     }
     return 0;
@@ -116,15 +116,15 @@ static void on_connection(uv_stream_t *server, int status) {
     data->accept_status = status;
     uv_unref((uv_handle_t *)server);
 }
-static void socket_bind(MVMThreadContext *tc, MVMOSHandle *h, MVMString *host, MVMint64 port) {
+static void socket_bind(MVMThreadContext *tc, MVMOSHandle *h, MVMString *host, MVMint64 port, MVMint32 backlog) {
     MVMIOSyncSocketData *data = (MVMIOSyncSocketData *)h->body.data;
     if (!data->ss.handle) {
         struct sockaddr *dest    = MVM_io_resolve_host_name(tc, host, port);
         uv_tcp_t        *socket  = MVM_malloc(sizeof(uv_tcp_t));
         int r;
 
-        if ((r = uv_tcp_init(tc->loop, socket)) < 0 ||
-                (r = uv_tcp_bind(socket, dest, 0)) < 0) {
+        if ((r = uv_tcp_init(tc->loop, socket)) != 0 ||
+                (r = uv_tcp_bind(socket, dest, 0)) != 0) {
             MVM_free(socket);
             MVM_free(dest);
             MVM_exception_throw_adhoc(tc, "Failed to bind: %s", uv_strerror(r));
@@ -134,7 +134,10 @@ static void socket_bind(MVMThreadContext *tc, MVMOSHandle *h, MVMString *host, M
         /* Start listening, but unref the socket so it won't get in the way of
          * other things we want to do on this event loop. */
         socket->data = data;
-        uv_listen((uv_stream_t *)socket, 1, on_connection);
+        if ((r = uv_listen((uv_stream_t *)socket, backlog, on_connection)) != 0) {
+            MVM_free(socket);
+            MVM_exception_throw_adhoc(tc, "Failed to listen: %s", uv_strerror(r));
+        }
         uv_unref((uv_handle_t *)socket);
 
         data->ss.handle = (uv_stream_t *)socket;
@@ -171,11 +174,13 @@ static const MVMIOOps op_table = {
     &sync_writable,
     NULL,
     NULL,
+    NULL,
     &seekable,
     &sockety,
     NULL,
     NULL,
     NULL,
+    NULL,
     gc_free
 };
 
@@ -187,7 +192,9 @@ static MVMObject * socket_accept(MVMThreadContext *tc, MVMOSHandle *h) {
             MVM_exception_throw_adhoc(tc, "Tried to accept() on a socket from outside its originating thread");
         }
         uv_ref((uv_handle_t *)data->ss.handle);
+        MVM_gc_mark_thread_blocked(tc);
         uv_run(tc->loop, UV_RUN_DEFAULT);
+        MVM_gc_mark_thread_unblocked(tc);
     }
 
     /* Check the accept worked out. */
@@ -205,7 +212,7 @@ static MVMObject * socket_accept(MVMThreadContext *tc, MVMOSHandle *h) {
             MVMIOSyncSocketData * const data   = MVM_calloc(1, sizeof(MVMIOSyncSocketData));
             data->ss.handle   = (uv_stream_t *)client;
             data->ss.encoding = MVM_encoding_type_utf8;
-            data->ss.sep      = '\n';
+            MVM_string_decode_stream_sep_default(tc, &(data->ss.sep_spec));
             result->body.ops  = &op_table;
             result->body.data = data;
             return (MVMObject *)result;
@@ -223,7 +230,8 @@ MVMObject * MVM_io_socket_create(MVMThreadContext *tc, MVMint64 listen) {
     MVMIOSyncSocketData * const data   = MVM_calloc(1, sizeof(MVMIOSyncSocketData));
     data->ss.handle   = NULL;
     data->ss.encoding = MVM_encoding_type_utf8;
-    data->ss.sep      = '\n';
+    data->ss.translate_newlines = 0;
+    MVM_string_decode_stream_sep_default(tc, &(data->ss.sep_spec));
     result->body.ops  = &op_table;
     result->body.data = data;
     return (MVMObject *)result;
