@@ -75,7 +75,9 @@ static MVMint32 read_to_buffer(MVMThreadContext *tc, MVMIOSyncStreamData *data,
         if (tc->loop != data->handle->loop) {
             MVM_exception_throw_adhoc(tc, "Tried to read() on a socket from outside its originating thread");
         }
+        MVM_gc_mark_thread_blocked(tc);
         uv_run(tc->loop, UV_RUN_DEFAULT);
+        MVM_gc_mark_thread_unblocked(tc);
         return 1;
     }
     else {
@@ -250,7 +252,7 @@ static MVMint64 closefh(MVMThreadContext *tc, MVMOSHandle *h) {
         uv_close((uv_handle_t *)data->handle, NULL);
         data->handle = NULL;
         if (data->ds) {
-            MVM_string_decodestream_destory(tc, data->ds);
+            MVM_string_decodestream_destroy(tc, data->ds);
             data->ds = NULL;
         }
     }
@@ -290,7 +292,7 @@ static void gc_free(MVMThreadContext *tc, MVMObject *h, void *d) {
             data->handle = NULL;
         }
         if (data->ds) {
-            MVM_string_decodestream_destory(tc, data->ds);
+            MVM_string_decodestream_destroy(tc, data->ds);
             data->ds = NULL;
         }
         MVM_string_decode_stream_sep_destroy(tc, &(data->sep_spec));
