@@ -10,6 +10,9 @@
  * such as ASCII and Latin-1, are normalized by definition).
  */
 
+#define DECODE_NOT_EOF  0
+#define DECODE_EOF      1
+
 /* Creates a new decoding stream. */
 MVMDecodeStream * MVM_string_decodestream_create(MVMThreadContext *tc, MVMint32 encoding,
         MVMint64 abs_byte_pos, MVMint32 translate_newlines) {
@@ -83,29 +86,38 @@ void MVM_string_decodestream_discard_to(MVMThreadContext *tc, MVMDecodeStream *d
 
 /* Does a decode run, selected by encoding. Returns non-zero if we actually
  * decoded more chars. */
-static MVMint32 run_decode(MVMThreadContext *tc, MVMDecodeStream *ds, const MVMint32 *stopper_chars, MVMDecodeStreamSeparators *sep_spec) {
+#define RUN_DECODE_NOTHING_DECODED          0
+#define RUN_DECODE_STOPPER_NOT_REACHED      1
+#define RUN_DECODE_STOPPER_REACHED          2
+static MVMuint32 run_decode(MVMThreadContext *tc, MVMDecodeStream *ds, const MVMint32 *stopper_chars, MVMDecodeStreamSeparators *sep_spec, MVMint32 eof) {
     MVMDecodeStreamChars *prev_chars_tail = ds->chars_tail;
+    MVMuint32 reached_stopper;
     switch (ds->encoding) {
     case MVM_encoding_type_utf8:
-        MVM_string_utf8_decodestream(tc, ds, stopper_chars, sep_spec);
+        reached_stopper = MVM_string_utf8_decodestream(tc, ds, stopper_chars, sep_spec);
         break;
     case MVM_encoding_type_ascii:
-        MVM_string_ascii_decodestream(tc, ds, stopper_chars, sep_spec);
+        reached_stopper = MVM_string_ascii_decodestream(tc, ds, stopper_chars, sep_spec);
         break;
     case MVM_encoding_type_latin1:
-        MVM_string_latin1_decodestream(tc, ds, stopper_chars, sep_spec);
+        reached_stopper = MVM_string_latin1_decodestream(tc, ds, stopper_chars, sep_spec);
         break;
     case MVM_encoding_type_windows1252:
-        MVM_string_windows1252_decodestream(tc, ds, stopper_chars, sep_spec);
+        reached_stopper = MVM_string_windows1252_decodestream(tc, ds, stopper_chars, sep_spec);
         break;
     case MVM_encoding_type_utf8_c8:
-        MVM_string_utf8_c8_decodestream(tc, ds, stopper_chars, sep_spec);
+        reached_stopper = MVM_string_utf8_c8_decodestream(tc, ds, stopper_chars, sep_spec, eof);
         break;
     default:
         MVM_exception_throw_adhoc(tc, "Streaming decode NYI for encoding %d",
             (int)ds->encoding);
     }
-    return ds->chars_tail != prev_chars_tail;
+    if (ds->chars_tail == prev_chars_tail)
+        return RUN_DECODE_NOTHING_DECODED;
+    else if (reached_stopper)
+        return RUN_DECODE_STOPPER_REACHED;
+    else
+        return RUN_DECODE_STOPPER_NOT_REACHED;
 }
 
 /* Gets the specified number of characters. If we are not yet able to decode
@@ -191,7 +203,7 @@ MVMString * MVM_string_decodestream_get_chars(MVMThreadContext *tc, MVMDecodeStr
     /* If we don't already have enough chars, try and decode more. */
     missing = missing_chars(tc, ds, chars);
     if (missing)
-        run_decode(tc, ds, &missing, NULL);
+        run_decode(tc, ds, &missing, NULL, DECODE_NOT_EOF);
 
     /* If we've got enough, assemble a string. Otherwise, give up. */
     if (missing_chars(tc, ds, chars) == 0)
@@ -201,8 +213,10 @@ MVMString * MVM_string_decodestream_get_chars(MVMThreadContext *tc, MVMDecodeStr
 }
 
 /* Gets characters up until one of the specified separators is encountered. If
- * we do not encounter it, returns 9. This may mean more input buffers are needed
- * or that we reached the end of the stream. */
+ * we do not encounter it, returns 0. This may mean more input buffers are needed
+ * or that we reached the end of the stream. Note that it assumes the separator
+ * will exist near the end of the buffer, if it occurs at all, due to decode
+ * streams looking for stoppers. */
 static MVMint32 have_separator(MVMThreadContext *tc, MVMDecodeStreamChars *start_chars, MVMint32 start_pos,
                                MVMDecodeStreamSeparators *sep_spec, MVMint32 sep_idx, MVMint32 sep_graph_pos) {
     MVMint32 sep_pos = 1;
@@ -227,6 +241,17 @@ static MVMint32 find_separator(MVMThreadContext *tc, const MVMDecodeStream *ds,
                                MVMDecodeStreamSeparators *sep_spec, MVMint32 *sep_length) {
     MVMint32 sep_loc = 0;
     MVMDecodeStreamChars *cur_chars = ds->chars_head;
+
+    /* First, skip over any buffers we need not consider. */
+    MVMint32 max_sep_chars = MVM_string_decode_stream_sep_max_chars(tc, sep_spec);
+    while (cur_chars && cur_chars->next) {
+        if (cur_chars->next->length < max_sep_chars)
+            break;
+        sep_loc += cur_chars->length;
+        cur_chars = cur_chars->next;
+    }
+
+    /* Now scan for the separator. */
     while (cur_chars) {
         MVMint32 start = cur_chars == ds->chars_head ? ds->chars_head_pos : 0;
         MVMint32 i, j;
@@ -263,9 +288,11 @@ MVMString * MVM_string_decodestream_get_until_sep(MVMThreadContext *tc, MVMDecod
      * the separator, so we may need to loop a few times around this. */
     sep_loc = find_separator(tc, ds, sep_spec, &sep_length);
     while (!sep_loc) {
-        if (!run_decode(tc, ds, NULL, sep_spec))
+        MVMuint32 decode_outcome = run_decode(tc, ds, NULL, sep_spec, DECODE_NOT_EOF);
+        if (decode_outcome == RUN_DECODE_NOTHING_DECODED)
             break;
-        sep_loc = find_separator(tc, ds, sep_spec, &sep_length);
+        if (decode_outcome == RUN_DECODE_STOPPER_REACHED)
+            sep_loc = find_separator(tc, ds, sep_spec, &sep_length);
     }
     if (sep_loc)
         return take_chars(tc, ds, sep_loc, chomp ? sep_length : 0);
@@ -278,7 +305,7 @@ MVMString * MVM_string_decodestream_get_until_sep(MVMThreadContext *tc, MVMDecod
 static void reached_eof(MVMThreadContext *tc, MVMDecodeStream *ds) {
     /* Decode all the things. */
     if (ds->bytes_head)
-        run_decode(tc, ds, NULL, NULL);
+        run_decode(tc, ds, NULL, NULL, DECODE_EOF);
 
     /* If there's some things left in the normalization buffer, take them. */
     MVM_unicode_normalizer_eof(tc, &(ds->norm));
@@ -312,15 +339,12 @@ MVMString * MVM_string_decodestream_get_until_sep_eof(MVMThreadContext *tc, MVMD
     return MVM_string_decodestream_get_all(tc, ds);
 }
 
-/* Decodes all the buffers, producing a string containing all the decoded
- * characters. */
-MVMString * MVM_string_decodestream_get_all(MVMThreadContext *tc, MVMDecodeStream *ds) {
+/* Produces a string consisting of the characters available now in all decdoed
+ * buffers. */
+static MVMString * get_all_in_buffer(MVMThreadContext *tc, MVMDecodeStream *ds) {
     MVMString *result = (MVMString *)MVM_repr_alloc_init(tc, tc->instance->VMString);
     result->body.storage_type = MVM_STRING_GRAPHEME_32;
 
-    /* Decode anything remaining and flush normalization buffer. */
-    reached_eof(tc, ds);
-
     /* If there's no codepoint buffer, then return the empty string. */
     if (!ds->chars_head) {
         result->body.storage.blob_32 = NULL;
@@ -382,6 +406,22 @@ MVMString * MVM_string_decodestream_get_all(MVMThreadContext *tc, MVMDecodeStrea
     return result;
 }
 
+/* Decodes all the buffers, signals EOF to flush any normalization buffers, and
+ * returns a string of all decoded chars. */
+MVMString * MVM_string_decodestream_get_all(MVMThreadContext *tc, MVMDecodeStream *ds) {
+    reached_eof(tc, ds);
+    return get_all_in_buffer(tc, ds);
+}
+
+/* Decodes all the buffers we have, and returns a string of all decoded chars.
+ * There may still be more to read after this, due to incomplete multi-byte
+ * or multi-codepoint sequences that are not yet completely processed. */
+MVMString * MVM_string_decodestream_get_available(MVMThreadContext *tc, MVMDecodeStream *ds) {
+    if (ds->bytes_head)
+        run_decode(tc, ds, NULL, NULL, DECODE_NOT_EOF);
+    return get_all_in_buffer(tc, ds);
+}
+
 /* Checks if we have the number of bytes requested. */
 MVMint64 MVM_string_decodestream_have_bytes(MVMThreadContext *tc, const MVMDecodeStream *ds, MVMint32 bytes) {
     MVMDecodeStreamBytes *cur_bytes = ds->bytes_head;
@@ -397,6 +437,19 @@ MVMint64 MVM_string_decodestream_have_bytes(MVMThreadContext *tc, const MVMDecod
     return 0;
 }
 
+/* Gets the number of bytes available. */
+MVMint64 MVM_string_decodestream_bytes_available(MVMThreadContext *tc, const MVMDecodeStream *ds) {
+    MVMDecodeStreamBytes *cur_bytes = ds->bytes_head;
+    MVMint32 available = 0;
+    while (cur_bytes) {
+        available += cur_bytes == ds->bytes_head
+            ? cur_bytes->length - ds->bytes_head_pos
+            : cur_bytes->length;
+        cur_bytes = cur_bytes->next;
+    }
+    return available;
+}
+
 /* Copies up to the requested number of bytes into the supplied buffer, and
  * returns the number of bytes we actually copied. Takes from from the start
  * of the stream. */
@@ -442,20 +495,28 @@ MVMint64 MVM_string_decodestream_tell_bytes(MVMThreadContext *tc, const MVMDecod
 
 /* Checks if the decode stream is empty. */
 MVMint32 MVM_string_decodestream_is_empty(MVMThreadContext *tc, MVMDecodeStream *ds) {
-    return !(ds->bytes_head || ds->chars_head || MVM_unicode_normalizer_available(tc, &(ds->norm)));
+    return !ds->bytes_head && !ds->chars_head && MVM_unicode_normalizer_empty(tc, &(ds->norm));
 }
 
 /* Destroys a decoding stream, freeing all associated memory (including the
  * buffers). */
-void MVM_string_decodestream_destory(MVMThreadContext *tc, MVMDecodeStream *ds) {
+void MVM_string_decodestream_destroy(MVMThreadContext *tc, MVMDecodeStream *ds) {
     MVMDecodeStreamBytes *cur_bytes = ds->bytes_head;
+    MVMDecodeStreamChars *cur_chars = ds->chars_head;
     while (cur_bytes) {
         MVMDecodeStreamBytes *next_bytes = cur_bytes->next;
         MVM_free(cur_bytes->bytes);
         MVM_free(cur_bytes);
         cur_bytes = next_bytes;
     }
+    while (cur_chars) {
+        MVMDecodeStreamChars *next_chars = cur_chars->next;
+        MVM_free(cur_chars->chars);
+        MVM_free(cur_chars);
+        cur_chars = next_chars;
+    }
     MVM_unicode_normalizer_cleanup(tc, &(ds->norm));
+    MVM_free(ds->decoder_state);
     MVM_free(ds);
 }
 
@@ -504,6 +565,16 @@ void MVM_string_decode_stream_sep_from_strings(MVMThreadContext *tc, MVMDecodeSt
     }
 }
 
+/* Returns the maximum length of any separator, in graphemes. */
+MVMint32 MVM_string_decode_stream_sep_max_chars(MVMThreadContext *tc, MVMDecodeStreamSeparators *sep_spec) {
+    MVMint32 i;
+    MVMint32 max_length = 1;
+    for (i = 0; i < sep_spec->num_seps; i++)
+        if (sep_spec->sep_lengths[i] > max_length)
+            max_length = sep_spec->sep_lengths[i];
+    return max_length;
+}
+
 /* Cleans up memory associated with a stream separator set. */
 void MVM_string_decode_stream_sep_destroy(MVMThreadContext *tc, MVMDecodeStreamSeparators *sep_spec) {
     MVM_free(sep_spec->sep_lengths);
