@@ -1,4 +1,8 @@
 #include "moar.h"
+#define UNI_CP_MALE_SIGN             0x2642
+#define UNI_CP_FEMALE_SIGN           0x2640
+#define UNI_CP_ZERO_WIDTH_JOINER     0x200D
+#define UNI_CP_ZERO_WIDTH_NON_JOINER 0x200C
 
 /* Maps outside-world normalization form codes to our internal set, validating
  * that we got something valid. */
@@ -77,36 +81,27 @@ void MVM_unicode_normalize_codepoints(MVMThreadContext *tc, const MVMObject *in,
     ((MVMArray *)out)->body.start     = 0;
     ((MVMArray *)out)->body.elems     = result_pos;
 }
-
-/* Takes an object, which must be of VMArray representation and holding
- * 32-bit integers. Treats them as Unicode codepoints, normalizes them at
- * Grapheme level, and returns the resulting NFG string. */
-MVMString * MVM_unicode_codepoints_to_nfg_string(MVMThreadContext *tc, const MVMObject *codes) {
+MVMString * MVM_unicode_codepoints_c_array_to_nfg_string(MVMThreadContext *tc, MVMCodepoint * cp_v, MVMint64 cp_count) {
     MVMNormalizer  norm;
-    MVMCodepoint  *input;
+    MVMint64       input_pos, result_pos, result_alloc;
     MVMGrapheme32 *result;
-    MVMint64       input_pos, input_codes, result_pos, result_alloc;
     MVMint32       ready;
     MVMString     *str;
 
-    /* Get input array; if it's empty, we're done already. */
-    assert_codepoint_array(tc, codes, "Code points to string input must be native array of 32-bit integers");
-    input       = (MVMCodepoint *)((MVMArray *)codes)->body.slots.u32 + ((MVMArray *)codes)->body.start;
-    input_codes = ((MVMArray *)codes)->body.elems;
-    if (input_codes == 0)
+    if (cp_count == 0)
         return tc->instance->str_consts.empty;
 
-    /* Guess output size based on input size. */
-    result_alloc = input_codes;
+    /* Guess output size based on cp_v size. */
+    result_alloc = cp_count;
     result       = MVM_malloc(result_alloc * sizeof(MVMCodepoint));
 
     /* Perform normalization at grapheme level. */
     MVM_unicode_normalizer_init(tc, &norm, MVM_NORMALIZE_NFG);
     input_pos  = 0;
     result_pos = 0;
-    while (input_pos < input_codes) {
+    while (input_pos < cp_count) {
         MVMGrapheme32 g;
-        ready = MVM_unicode_normalizer_process_codepoint_to_grapheme(tc, &norm, input[input_pos], &g);
+        ready = MVM_unicode_normalizer_process_codepoint_to_grapheme(tc, &norm, cp_v[input_pos], &g);
         if (ready) {
             maybe_grow_result(&result, &result_alloc, result_pos + ready);
             result[result_pos++] = g;
@@ -130,6 +125,20 @@ MVMString * MVM_unicode_codepoints_to_nfg_string(MVMThreadContext *tc, const MVM
     return str;
 }
 
+/* Takes an object, which must be of VMArray representation and holding
+ * 32-bit integers. Treats them as Unicode codepoints, normalizes them at
+ * Grapheme level, and returns the resulting NFG string. */
+MVMString * MVM_unicode_codepoints_to_nfg_string(MVMThreadContext *tc, const MVMObject *codes) {
+    MVMCodepoint  *input;
+    MVMint64       input_codes;
+
+    assert_codepoint_array(tc, codes, "Code points to string input must be native array of 32-bit integers");
+
+    input       = (MVMCodepoint *)((MVMArray *)codes)->body.slots.u32 + ((MVMArray *)codes)->body.start;
+    input_codes = ((MVMArray *)codes)->body.elems;
+    return MVM_unicode_codepoints_c_array_to_nfg_string(tc, input, input_codes);
+}
+
 /* Takes an NFG string and populates the array out, which must be a 32-bit
  * integer array, with codepoints normalized according to the specified
  * normalization form. */
@@ -281,13 +290,11 @@ static void decomp_codepoint_to_buffer(MVMThreadContext *tc, MVMNormalizer *n, M
     /* See if we actually need to decompose (can skip if the decomposition
      * type is None, or we're only doing Canonical decomposition and it is
      * anything except Canonical). */
-    const char *type = MVM_unicode_codepoint_get_property_cstr(tc, cp, MVM_UNICODE_PROPERTY_DECOMPOSITION_TYPE);
+    MVMint16 cp_DT = MVM_unicode_codepoint_get_property_int(tc, cp, MVM_UNICODE_PROPERTY_DECOMPOSITION_TYPE);
     MVMint64 decompose = 1;
-    if (!type)
+    if (cp_DT == MVM_UNICODE_PVALUE_DT_NONE)
         decompose = 0;
-    else if (strcmp(type, "None") == 0)
-        decompose = 0;
-    else if (!MVM_NORMALIZE_COMPAT_DECOMP(n->form) && strcmp(type, "Canonical") != 0)
+    else if (!MVM_NORMALIZE_COMPAT_DECOMP(n->form) && cp_DT != MVM_UNICODE_PVALUE_DT_CANONICAL )
         decompose = 0;
     if (decompose) {
         /* We need to decompose. Get the decomp spec and go over the things in
@@ -320,6 +327,13 @@ static MVMint64 passes_quickcheck(MVMThreadContext *tc, const MVMNormalizer *n,
     return pval && pval[0] == 'Y';
 }
 
+MVM_STATIC_INLINE MVMint32 fast_atoi( const char * dec_str ) {
+    MVMint32 value = 0;
+    while( *dec_str ) {
+        value = value*10 + (*dec_str++ - '0');
+    }
+    return value;
+}
 /* Gets the canonical combining class for a codepoint. */
 static MVMint64 ccc(MVMThreadContext *tc, MVMCodepoint cp) {
     if (cp < MVM_NORMALIZE_FIRST_NONZERO_CCC) {
@@ -327,7 +341,7 @@ static MVMint64 ccc(MVMThreadContext *tc, MVMCodepoint cp) {
     }
     else {
         const char *ccc_str = MVM_unicode_codepoint_get_property_cstr(tc, cp, MVM_UNICODE_PROPERTY_CANONICAL_COMBINING_CLASS);
-        return !ccc_str || strlen(ccc_str) > 3 ? 0 : atoi(ccc_str);
+        return !ccc_str || strlen(ccc_str) > 3 ? 0 : fast_atoi(ccc_str);
     }
 }
 
@@ -337,25 +351,28 @@ static MVMint64 ccc(MVMThreadContext *tc, MVMCodepoint cp) {
  * a fast path. */
 static MVMint32 is_control_beyond_latin1(MVMThreadContext *tc, MVMCodepoint in) {
     /* U+200C ZERO WIDTH NON-JOINER and U+200D ZERO WIDTH JOINER are excluded. */
-    if (in != 0x200C && in != 0x200D) {
+    if (in != UNI_CP_ZERO_WIDTH_NON_JOINER && in != UNI_CP_ZERO_WIDTH_JOINER) {
         /* Consider general property. */
         const char *genprop = MVM_unicode_codepoint_get_property_cstr(tc, in,
             MVM_UNICODE_PROPERTY_GENERAL_CATEGORY);
-        if (genprop[0] == 'Z') {
-            /* Line_Separator and Paragraph_Separator are controls. */
-            return genprop[1] == 'l' || genprop[1] == 'p';
-        }
-        if (genprop[0] == 'C') {
-            /* Control, Surrogate, and Format are controls. */
-            if (genprop[1] == 'c' || genprop[1] == 's' || genprop[1] == 'f') {
-                return 1;
-            }
-
-            /* Unassigned is, but only for Default_Ignorable_Code_Point. */
-            if (genprop[1] == 'n') {
-                return MVM_unicode_codepoint_get_property_int(tc, in,
-                    MVM_UNICODE_PROPERTY_DEFAULT_IGNORABLE_CODE_POINT) != 0;
-            }
+        switch (genprop[0]) {
+            case 'Z':
+                /* Line_Separator and Paragraph_Separator are controls. */
+                return genprop[1] == 'l' || genprop[1] == 'p';
+            case 'C':
+                /* Control, Surrogate are controls. */
+                if (genprop[1] == 'c' || genprop[1] == 's') {
+                    return 1;
+                }
+                if (genprop[1] == 'f' ) {
+                    /* Format can have special properties (not control) */
+                    return 0;
+                }
+                /* Unassigned is, but only for Default_Ignorable_Code_Point. */
+                if (genprop[1] == 'n') {
+                    return MVM_unicode_codepoint_get_property_int(tc, in,
+                        MVM_UNICODE_PROPERTY_DEFAULT_IGNORABLE_CODE_POINT) != 0;
+                }
         }
     }
     return 0;
@@ -491,6 +508,11 @@ static MVMint32 is_grapheme_extend(MVMThreadContext *tc, MVMCodepoint cp) {
     return MVM_unicode_codepoint_get_property_int(tc, cp,
         MVM_UNICODE_PROPERTY_GRAPHEME_EXTEND);
 }
+static MVMint32 is_grapheme_prepend(MVMThreadContext *tc, MVMCodepoint cp) {
+    return MVM_unicode_codepoint_get_property_int(tc, cp,
+        MVM_UNICODE_PROPERTY_PREPENDED_CONCATENATION_MARK);
+}
+
 static MVMint32 is_spacing_mark(MVMThreadContext *tc, MVMCodepoint cp) {
     const char *genprop = MVM_unicode_codepoint_get_property_cstr(tc, cp,
         MVM_UNICODE_PROPERTY_GENERAL_CATEGORY);
@@ -507,6 +529,8 @@ static MVMint32 is_spacing_mark(MVMThreadContext *tc, MVMCodepoint cp) {
     }
 }
 static MVMint32 should_break(MVMThreadContext *tc, MVMCodepoint a, MVMCodepoint b) {
+    int GCB_a = MVM_unicode_codepoint_get_property_int(tc, a, MVM_UNICODE_PROPERTY_GRAPHEME_CLUSTER_BREAK);
+    int GCB_b = MVM_unicode_codepoint_get_property_int(tc, b, MVM_UNICODE_PROPERTY_GRAPHEME_CLUSTER_BREAK);
     /* Don't break between \r and \n, but otherwise break around \r. */
     if (a == 0x0D && b == 0x0A)
         return 0;
@@ -528,19 +552,64 @@ static MVMint32 should_break(MVMThreadContext *tc, MVMCodepoint a, MVMCodepoint
             return !(strcmp(hst_b, "T") == 0);
     }
 
-    /* Don't break between regional indicators. */
-    if (is_regional_indicator(a) && is_regional_indicator(b))
-        return 0;
-
-    /* Don't break before extenders. */
-    if (is_grapheme_extend(tc, b))
-        return 0;
-
-    /* Don't break before spacing marks. (In the Unicode version at the time
-     * of implementing, there were no Prepend characters, so we don't worry
-     * about that rule for now). */
-    if (is_spacing_mark(tc, b))
-        return 0;
+    switch (GCB_a) {
+        case MVM_UNICODE_PVALUE_GCB_REGIONAL_INDICATOR:
+            if ( GCB_b == MVM_UNICODE_PVALUE_GCB_REGIONAL_INDICATOR )
+                return 0;
+            break;
+        /* Don't break after Prepend Grapheme_Cluster_Break=Prepend */
+        case MVM_UNICODE_PVALUE_GCB_PREPEND:
+            // If it's a control character remember to break
+            if (is_control_beyond_latin1(tc, b )) {
+                return 1;
+            }
+            // Otherwise don't break
+            return 0;
+        /* Don't break after ZWJ for E_Base_GAZ or Glue_After_ZWJ */
+        case MVM_UNICODE_PVALUE_GCB_ZWJ:
+            switch (GCB_b) {
+                case MVM_UNICODE_PVALUE_GCB_E_BASE_GAZ:
+                case MVM_UNICODE_PVALUE_GCB_ZWJ:
+                case MVM_UNICODE_PVALUE_GCB_GLUE_AFTER_ZWJ:
+                    return 0;
+            }
+            if ( b == UNI_CP_FEMALE_SIGN || b == UNI_CP_MALE_SIGN )
+                return 0;
+        case MVM_UNICODE_PVALUE_GCB_E_MODIFIER:
+            if (MVM_unicode_codepoint_get_property_int(tc, b, MVM_UNICODE_PROPERTY_EMOJI_MODIFIER_BASE)) {
+                /* Don't break after ZWJ if it's an Emoji Sequence.
+                 * At the moment FEMALE SIGN and MALE SIGN don't have different
+                 * GCB properties, or any special Emoji properties (Unicode 9.0),
+                 * so we explictly check these codepoints here */
+                if ( b == UNI_CP_FEMALE_SIGN || b == UNI_CP_MALE_SIGN )
+                    return 0;
+            }
+            break;
+    }
+    switch (GCB_b) {
+        /* Don't break before extending chars */
+        case MVM_UNICODE_PVALUE_GCB_EXTEND:
+            return 0;
+        /* Don't break before ZWJ */
+        case MVM_UNICODE_PVALUE_GCB_ZWJ:
+            return 0;
+        case MVM_UNICODE_PVALUE_GCB_E_MODIFIER:
+            switch (GCB_a) {
+                case MVM_UNICODE_PVALUE_GCB_E_BASE_GAZ:
+                    return 0;
+                case MVM_UNICODE_PVALUE_GCB_E_BASE:
+                    return 0;
+                /* Don't break
+                 * when in Emoji Sequences
+                 * we don't save state so can't support this now
+                 *case MVM_UNICODE_PVALUE_GCB_EXTEND:
+                 *    return 0; */
+            }
+            break;
+        /* Don't break before spacing marks. */
+        case MVM_UNICODE_PVALUE_GCB_SPACINGMARK:
+            return 0;
+    }
 
     /* Otherwise break. */
     return 1;
@@ -578,11 +647,12 @@ static void grapheme_composition(MVMThreadContext *tc, MVMNormalizer *n, MVMint3
  * compute the normalization. */
 MVMint32 MVM_unicode_normalizer_process_codepoint_full(MVMThreadContext *tc, MVMNormalizer *n, MVMCodepoint in, MVMCodepoint *out) {
     MVMint64 qc_in, ccc_in;
-
+    int is_prepend = is_grapheme_prepend(tc, in);
     /* If it's a control character (outside of the range we checked in the
      * fast path) then it's a normalization terminator. */
-    if (in > 0xFF && is_control_beyond_latin1(tc, in))
+    if (in > 0xFF && is_control_beyond_latin1(tc, in) && !is_prepend) {
         return MVM_unicode_normalizer_process_codepoint_norm_terminator(tc, n, in, out);
+    }
 
     /* Do a quickcheck on the codepoint we got in and get its CCC. */
     qc_in  = passes_quickcheck(tc, n, in);
@@ -623,7 +693,7 @@ MVMint32 MVM_unicode_normalizer_process_codepoint_full(MVMThreadContext *tc, MVM
          * buffer, if any. We need to do this since it may have passed
          * quickcheck, but having seen some character that does pass then we
          * must make sure we decomposed the prior passing one too. */
-        if (MVM_NORMALIZE_COMPOSE(n->form) && n->buffer_end != n->buffer_norm_end) {
+        if (MVM_NORMALIZE_COMPOSE(n->form) && n->buffer_end != n->buffer_norm_end && !is_prepend) {
             MVMCodepoint decomp = n->buffer[n->buffer_end - 1];
             n->buffer_end--;
             decomp_codepoint_to_buffer(tc, n, decomp);
