@@ -1,13 +1,119 @@
+/* Compares two strings, based on Unicode Collation Algorithm
+ *    0   The strings are identical, includes codepoints
+ * -1/1   We used the primary collation values to decide the result
+ * -2/2   We used the secondary, meaning the primary weights were equal
+ * -3/3   We used the tetriary, meaning the primary and also the secondary
+          weights (if we used that option, which is the default) were equal.
+ *   -4/4 We used codepoints to decide because primary, secondary and/or tetriary
+          were equal (depending on if secondary and tetriary were requested).
+          If the codepoints are all the same, we will decide based on length.
+ * -10/10 The collation algorithm was not able to be applied and so they
+          were compared based on codepoints. If codepoints were equal they were
+          compared by length. */
+MVMint32 MVM_unicode_collation_primary (MVMThreadContext *tc, MVMint32 codepoint) {
+     return MVM_unicode_codepoint_get_property_int(tc, codepoint, MVM_UNICODE_PROPERTY_MVM_COLLATION_PRIMARY);
+}
+MVMint32 MVM_unicode_collation_secondary (MVMThreadContext *tc, MVMint32 codepoint) {
+     return MVM_unicode_codepoint_get_property_int(tc, codepoint, MVM_UNICODE_PROPERTY_MVM_COLLATION_SECONDARY);
+}
+MVMint32 MVM_unicode_collation_tertiary (MVMThreadContext *tc, MVMint32 codepoint) {
+     return MVM_unicode_codepoint_get_property_int(tc, codepoint, MVM_UNICODE_PROPERTY_MVM_COLLATION_TERTIARY);
+}
+MVMint64 MVM_unicode_string_compare
+        (MVMThreadContext *tc, MVMString *a, MVMString *b,
+         MVMint32 collation_mode, MVMint32 lang_mode, MVMint32 country_mode) {
+    MVMStringIndex alen, blen, i, scanlen;
+    /* Iteration variables */
+    MVMGrapheme32 ai;
+    MVMGrapheme32 bi;
+    /* Collation order numbers */
+    MVMint32 ai_coll_val;
+    MVMint32 bi_coll_val;
+    MVM_string_check_arg(tc, a, "compare");
+    MVM_string_check_arg(tc, b, "compare");
+    /* Simple cases when one or both are zero length. */
+    alen = MVM_string_graphs(tc, a);
+    blen = MVM_string_graphs(tc, b);
+    if (alen == 0)
+        return blen == 0 ? 0 : -1;
+    if (blen == 0)
+        return 1;
+
+    /* Otherwise, need to scan them. */
+    scanlen = alen > blen ? blen : alen;
+    for (i = 0; i < scanlen; i++) {
+        ai = MVM_string_get_grapheme_at_nocheck(tc, a, i);
+        bi = MVM_string_get_grapheme_at_nocheck(tc, b, i);
+        /* If they are the same grapheme */
+        if (ai != bi) {
+            /* only try and get a property if the value it is isn't a synthetic
+             * grapheme. we should change from get_grapheme to something to get the
+             * NFC form */
+            if ( ai >= 0 || bi >= 0 ) {
+                /* Get the primary collation value for the grapheme */
+                ai_coll_val = MVM_unicode_collation_primary(tc, ai);
+                bi_coll_val = MVM_unicode_collation_primary(tc, bi);
+                /* If we don't find a collation value,
+                   we should compare by codepoint */
+                /* Eventually we should try and catch codepoints that don't
+                   have a collation value. We would then need to decompose it and
+                   apply weighting based on their decomposed values. */
+                if (ai_coll_val == 0 || bi_coll_val == 0) {
+                    /* return -10 or 10 to indicate we didn't use the collation
+                       algorithm */
+                    return ai < bi ? -10 :
+                           ai > bi ?  10 :
+                                       0 ;
+                }
+
+                /* If both have primary collation values ( they are not 0 ) */
+                if ( (ai_coll_val != 0 && bi_coll_val != 0) && (ai_coll_val != bi_coll_val) ) {
+                    return ai_coll_val < bi_coll_val ? -1 : 1;
+                }
+                /* If both have the same primary collation values */
+                ai_coll_val += MVM_unicode_collation_secondary(tc, ai);
+                bi_coll_val += MVM_unicode_collation_secondary(tc, bi);
+                if ( (ai_coll_val != 0 && bi_coll_val != 0) && (ai_coll_val != bi_coll_val) ) {
+                    return ai_coll_val < bi_coll_val ? -2 : 2;
+                }
+                /* If both have the same tertiary collation values */
+                ai_coll_val = MVM_unicode_collation_tertiary(tc, ai);
+                bi_coll_val = MVM_unicode_collation_tertiary(tc, bi);
+                if ( (ai_coll_val != 0 && bi_coll_val != 0) && (ai_coll_val != bi_coll_val) ) {
+                    return ai_coll_val < bi_coll_val ? -3 : 3;
+                }
+                /* All the collation values were equal. Check codepoints */
+                return ai < bi ? -4 :
+                       ai > bi ?  4 :
+                                  0 ;
+
+            }
+            /* For now, if it's a synthetic codepoint just compare by codepoint. */
+            else {
+                return ai < bi ? -10 :
+                       ai > bi ?  10 :
+                                   1 ;
+            }
+        }
+    }
+
+    /* All shared chars equal, so go on length. */
+    return alen < blen ? -1 :
+           alen > blen ?  1 :
+                          0 ;
+}
 
 /* Looks up a codepoint by name. Lazily constructs a hash. */
 MVMGrapheme32 MVM_unicode_lookup_by_name(MVMThreadContext *tc, MVMString *name) {
     MVMuint64 size;
     char *cname = MVM_string_ascii_encode(tc, name, &size, 0);
+    size_t cname_len = strlen((const char *) cname );
     MVMUnicodeNameRegistry *result;
     if (!codepoints_by_name) {
         generate_codepoints_by_name(tc);
     }
-    HASH_FIND(hash_handle, codepoints_by_name, cname, strlen((const char *)cname), result);
+    HASH_FIND(hash_handle, codepoints_by_name, cname, cname_len, result);
+
     MVM_free(cname);
     return result ? result->codepoint : -1;
 }
@@ -130,6 +236,7 @@ MVMuint32 MVM_unicode_get_case_change(MVMThreadContext *tc, MVMCodepoint codepoi
 
 /* XXX make all the statics members of the global MVM instance instead? */
 static MVMUnicodeNameRegistry *property_codes_by_names_aliases;
+static MVMUnicodeGraphemeNameRegistry *property_codes_by_seq_names;
 
 static void generate_property_codes_by_names_aliases(MVMThreadContext *tc) {
     MVMuint32 num_names = num_unicode_property_keypairs;
@@ -142,6 +249,17 @@ static void generate_property_codes_by_names_aliases(MVMThreadContext *tc) {
             entry->name, strlen(entry->name), entry);
     }
 }
+static void generate_property_codes_by_seq_names(MVMThreadContext *tc) {
+    MVMuint32 num_names = num_unicode_seq_keypairs;
+
+    while (num_names--) {
+        MVMUnicodeGraphemeNameRegistry *entry = MVM_malloc(sizeof(MVMUnicodeGraphemeNameRegistry));
+        entry->name = (char *)uni_seq_pairs[num_names].name;
+        entry->structindex = uni_seq_pairs[num_names].value;
+        HASH_ADD_KEYPTR(hash_handle, property_codes_by_seq_names,
+            entry->name, strlen(entry->name), entry);
+    }
+}
 
 MVMint32 MVM_unicode_name_to_property_code(MVMThreadContext *tc, MVMString *name) {
     MVMuint64 size;
@@ -156,7 +274,7 @@ MVMint32 MVM_unicode_name_to_property_code(MVMThreadContext *tc, MVMString *name
 }
 
 static void generate_unicode_property_values_hashes(MVMThreadContext *tc) {
-    MVMUnicodeNameRegistry **hash_array = MVM_calloc(sizeof(MVMUnicodeNameRegistry *), MVM_NUM_PROPERTY_CODES);
+    MVMUnicodeNameRegistry **hash_array = MVM_calloc(MVM_NUM_PROPERTY_CODES, sizeof(MVMUnicodeNameRegistry *));
     MVMuint32 index = 0;
     MVMUnicodeNameRegistry *entry = NULL, *binaries = NULL;
     for ( ; index < num_unicode_property_value_keypairs; index++) {
@@ -296,3 +414,34 @@ void MVM_unicode_release(MVMThreadContext *tc)
     }
     uv_mutex_unlock(&property_hash_count_mutex);
 }
+/* Looks up a grapheme by name. Lazily constructs a hash. */
+MVMString * MVM_unicode_string_from_name(MVMThreadContext *tc, MVMString *name) {
+    MVMuint64 size;
+    char *cname = MVM_string_ascii_encode(tc, name, &size, 0);
+    MVMUnicodeGraphemeNameRegistry *result;
+
+    MVMGrapheme32 result_graph;
+    const MVMint32 * uni_seq;
+    int array_size;
+
+    result_graph = MVM_unicode_lookup_by_name(tc, name);
+    /* If it's just a codepoint, return that */
+    if (result_graph >= 0) {
+        MVM_free(cname);
+        return MVM_string_chr(tc, result_graph);
+    }
+    if (!property_codes_by_seq_names) {
+        generate_property_codes_by_seq_names(tc);
+    }
+    HASH_FIND(hash_handle, property_codes_by_seq_names, cname, strlen((const char *)cname), result);
+    MVM_free(cname);
+    /* If we can't find a result return an empty string */
+    if (!result)
+        return tc->instance->str_consts.empty;
+
+    uni_seq = uni_seq_enum[result->structindex];
+    /* The first element is the number of codepoints in the sequence */
+    array_size = uni_seq[0];
+    return MVM_unicode_codepoints_c_array_to_nfg_string(tc, (MVMCodepoint *) uni_seq + 1, array_size);
+
+}
